from time import time
import numpy as np
from numba import njit, prange
from numba.extending import overload
import numexpr as ne
from functools import partial
import dask.array as da
from uuid import uuid4
from pfb.utils.misc import norm_diff, JIT_OPTIONS
from pfb.utils.naming import xds_from_list
from ducc0.misc import empty_noncritical
iFs = np.fft.ifftshift
Fs = np.fft.fftshift

# from pfb.utils import logging as pfb_logging
# log = pfb_logging.get_logger('PCG')

@njit(nogil=True, cache=False, parallel=False)
def update(x, xp, r, rp, p, Ap, alpha):
    return update_impl(x, xp, r, rp, p, Ap, alpha)


def update_impl(x, xp, r, rp, p, Ap, alpha):
    return NotImplementedError


@overload(update_impl, jit_options={**JIT_OPTIONS})  #, "parallel":True})
def nb_update_impl(x, xp, r, rp, p, Ap, alpha):
    if x.ndim==3:
        def impl(x, xp, r, rp, p, Ap, alpha):
            nband, nx, ny = x.shape
            for b in range(nband):
                for i in prange(nx):
                    for j in range(ny):
                        x[b, i, j] = xp[b, i, j] + alpha * p[b, i, j]
                        r[b, i, j] = rp[b, i, j] + alpha * Ap[b, i, j]
            return x, r
    elif x.ndim==2:
        def impl(x, xp, r, rp, p, Ap, alpha):
            nx, ny = x.shape
            for i in prange(nx):
                for j in range(ny):
                    x[i, j] = xp[i, j] + alpha * p[i, j]
                    r[i, j] = rp[i, j] + alpha * Ap[i, j]
            return x, r
    else:
        raise ValueError("norm_diff is only implemented for 2D or 3D arrays")

    return impl


@njit(**JIT_OPTIONS, parallel=True)
def alpha_update(r, y, p, Ap):
    return alpha_update_impl(r, y, p, Ap)


def alpha_update_impl(r, y, p, Ap):
    return NotImplementedError


@overload(alpha_update_impl, jit_options={**JIT_OPTIONS, 'parallel':True})
def nb_alpha_update_impl(r, y, p, Ap):
    if r.ndim==2:
        def impl(r, y, p, Ap):
            rnorm = 0.0
            rnorm_den = 0.0
            nx, ny = r.shape
            for i in prange(nx):
                for j in range(ny):
                    rnorm += r[i,j]*y[i,j]
                    rnorm_den += p[i,j]*Ap[i,j]

            alpha = rnorm/rnorm_den
            return rnorm, alpha
    elif r.ndim==3:
        def impl(r, y, p, Ap):
            rnorm = 0.0
            rnorm_den = 0.0
            nband, nx, ny = r.shape
            for b in range(nband):
                for i in prange(nx):
                    for j in range(ny):
                        rnorm += r[b,i,j]*y[b,i,j]
                        rnorm_den += p[b,i,j]*Ap[b,i,j]

            alpha = rnorm/rnorm_den
            return rnorm, alpha
    return impl


@njit(**JIT_OPTIONS, parallel=True)
def beta_update(r, y, p, rnorm):
    return beta_update_impl(r, y, p, rnorm)


def beta_update_impl(r, y, p, rnorm):
    return NotImplementedError


@overload(beta_update_impl, jit_options={**JIT_OPTIONS, 'parallel':True})
def nb_beta_update_impl(r, y, p, rnorm):
    if r.ndim==2:
        def impl(r, y, p, rnorm):
            rnorm_next = 0.0
            nx, ny = r.shape
            for i in prange(nx):
                for j in range(ny):
                    rnorm_next += r[i,j]*y[i,j]

            beta = rnorm_next/rnorm

            for i in prange(nx):
                for j in range(ny):
                    p[i,j] = beta * p[i,j] - y[i,j]
            return rnorm_next, p
    elif r.ndim==3:
        def impl(r, y, p, rnorm):
            rnorm_next = 0.0
            nband, nx, ny = r.shape
            for b in range(nband):
                for i in prange(nx):
                    for j in range(ny):
                        rnorm_next += r[b,i,j]*y[b,i,j]

            beta = rnorm_next/rnorm
            for b in range(nband):
                for i in prange(nx):
                    for j in range(ny):
                        p[b,i,j] = beta * p[b,i,j] - y[b,i,j]
            return rnorm_next, p
    return impl


def pcg(A,
        b,
        x0=None,
        M=None,
        tol=1e-5,
        maxit=500,
        minit=100,
        verbosity=1,
        report_freq=10,
        backtrack=True,
        return_resid=False):

    if x0 is None:
        x0 = np.zeros(b.shape, dtype=b.dtype)

    if M is None:
        def M(x): return x

    r = A(x0) - b
    y = M(r)
    if not np.any(y):
        print(f"Initial residual is zero")
        return x0
    p = -y
    rnorm = np.vdot(r, y)
    if np.isnan(rnorm) or rnorm == 0.0:
        phi0 = 1.0
    else:
        phi0 = rnorm
    k = 0
    x = x0
    eps = 1.0
    stall_count = 0
    xp = x.copy()
    rp = r.copy()
    tcopy = 0.0
    tA = 0.0
    tvdot = 0.0
    tupdate = 0.0
    tp = 0.0
    tnorm = 0.0
    tii = time()
    while (eps > tol or k < minit) and k < maxit and stall_count < 5:
        ti = time()
        np.copyto(xp, x)
        np.copyto(rp, r)
        tcopy += (time() - ti)
        ti = time()
        Ap = A(p)
        tA += (time() - ti)
        ti = time()
        rnorm = np.vdot(r, y)
        alpha = rnorm / np.vdot(p, Ap)
        tvdot += (time() - ti)
        ti = time()
        # x = xp + alpha * p
        # r = rp + alpha * Ap
        ne.evaluate('xp + alpha*p',
                    out=x,
                    local_dict={
                        'xp': xp,
                        'alpha': alpha,
                        'p': p},
                    casting='unsafe')
        ne.evaluate('rp + alpha*Ap',
                    out=r,
                    local_dict={
                        'rp': rp,
                        'alpha': alpha,
                        'Ap': Ap},
                    casting='unsafe')
        # x, r = update(x, xp, r, rp, p, Ap, alpha)
        tupdate += (time() - ti)
        y = M(r)

        # while rnorm_next > rnorm and backtrack:  # TODO - better line search
        #     alpha *= 0.75
        #     x = xp + alpha * p
        #     r = rp + alpha * Ap
        #     y = M(r)
        #     rnorm_next = np.vdot(r, y)

        ti = time()
        rnorm_next = np.vdot(r, y)
        beta = rnorm_next / rnorm
        ne.evaluate('beta*p-y',
                    out=p,
                    local_dict={
                        'beta': beta,
                        'p': p,
                        'y': y},
                    casting='unsafe')
        # p = beta * p - y
        # p *= beta
        # p -= y
        # rnorm, p = beta_update(r, y, p, rnorm)
        tp += (time() - ti)
        rnorm = rnorm_next
        k += 1
        epsp = eps
        ti = time()
        eps = norm_diff(x, xp)
        phi = rnorm / phi0
        tnorm += (time() - ti)

        if np.abs(epsp - eps) < 1e-3*tol:
            stall_count += 1

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} eps = {eps:.3e}, phi = {phi:.3e}")
    ttot = time() - tii
    ttally = tcopy + tA + tvdot + tupdate + tp + tnorm
    if verbosity > 1:
        print('tcopy = ', tcopy/ttot)
        print('tA = ', tA/ttot)
        print('tvdot = ', tvdot/ttot)
        print('tupdate = ', tupdate/ttot)
        print('tp = ', tp/ttot)
        print('tnorm = ', tnorm/ttot)
        print('ttally = ', ttally/ttot)

    if k >= maxit:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}")
    elif stall_count >= 5:
        if verbosity:
            print(f"Stalled after {k} iterations with eps = {eps:.3e}")
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations")
    if not return_resid:
        return x
    else:
        return x, r


def _pcg_psf_impl(psfhat,
                  b,
                  x0,
                  beam,
                  lastsize,
                  nthreads,
                  eta,
                  tol=1e-5,
                  maxit=500,
                  minit=100,
                  verbosity=1,
                  report_freq=10,
                  backtrack=True):
    '''
    A specialised distributed version of pcg when the operator implements
    convolution with the psf (+ L2 regularisation by sigma**2)
    '''
    nband, nx, ny = b.shape
    _, nx_psf, nyo2 = psfhat.shape
    model = np.zeros((nband, nx, ny), dtype=b.dtype, order='C')
    # PCG preconditioner
    if eta > 0:
        def M(x): return x / eta
    else:
        M = None
    from pfb.operators.hessian import _hessian_psf_slice
    for k in range(nband):
        xpad = empty_noncritical((nx_psf, lastsize), dtype=b.dtype)
        xhat = empty_noncritical((nx_psf, nyo2), dtype=psfhat.dtype)
        xout = empty_noncritical((nx, ny), dtype=b.dtype)
        A = partial(_hessian_psf_slice,
                    xpad=xpad,
                    xhat=xhat,
                    xout=xout,
                    abspsf=np.abs(psfhat[k]),
                    beam=beam[k],
                    lastsize=lastsize,
                    nthreads=nthreads,
                    eta=eta[k])

        model[k] = pcg(A, b[k], x0[k],
                       M=M, tol=tol, maxit=maxit, minit=minit,
                       verbosity=verbosity, report_freq=report_freq,
                       backtrack=backtrack)

    return model

def _pcg_psf(psfhat,
             b,
             x0,
             beam,
             lastsize,
             nthreads,
             eta,
             cgopts):
    return _pcg_psf_impl(psfhat,
                         b,
                         x0,
                         beam,
                         lastsize,
                         nthreads,
                         eta,
                         **cgopts)

def pcg_psf(psfhat,
            b,
            x0,
            beam,
            lastsize,
            nthreads,
            eta,
            cgopts,
            compute=True):

    if not isinstance(x0, da.Array):
        x0 = da.from_array(x0, chunks=(1, -1, -1),
                          name="x-" + uuid4().hex)

    if not isinstance(psfhat, da.Array):
        psfhat = da.from_array(psfhat, chunks=(1, -1, -1),
                               name="psfhat-" + uuid4().hex)
    if not isinstance(b, da.Array):
        b = da.from_array(b, chunks=(1, -1, -1),
                          name="psfhat-" + uuid4().hex)
        
    nband = b.shape[0]
    if isinstance(eta, float):
        eta = np.tile(eta, nband)
    else:
        eta = np.array(eta)
        assert eta.size == nband
    eta = da.from_array(eta, chunks=(1,),
                        name="eta-" + uuid4().hex)

    if beam is None:
        bout = None
    else:
        bout = ('nb', 'nx', 'ny')
        if beam.ndim == 2:
            beam = beam[None]

        if not isinstance(beam, da.Array):
            beam = da.from_array(beam, chunks=(1, -1, -1),
                                 name="beam-" + uuid4().hex)
        if beam.shape[0] == 1:
            beam = da.tile(beam, (psfhat.shape[0], 1, 1))
        elif beam.shape[0] != psfhat.shape[0]:
            raise ValueError('Beam has incorrect shape')

    model = da.blockwise(_pcg_psf, ('nb', 'nx', 'ny'),
                         psfhat, ('nb', 'nx', 'ny'),
                         b, ('nb', 'nx', 'ny'),
                         x0, ('nb', 'nx', 'ny'),
                         beam, bout,
                         lastsize, None,
                         nthreads, None,
                         eta, ('nb',),
                         cgopts, None,
                         align_arrays=False,
                         dtype=b.dtype)
    if compute:
        return model.compute()
    else:
        return model


def pcg_dds(ds_name,
            eta,  # regularisation for Hessian approximation
            mask=1.0,
            use_psf=True,
            residual_name='RESIDUAL',
            model_name='MODEL',
            do_wgridding=True,
            epsilon=5e-4,
            double_accum=True,
            nthreads=1,
            zero_model_outside_mask=False,
            tol=1e-5,
            maxit=500,
            verbosity=1,
            report_freq=10):
    '''
    pcg for fluxtractor
    '''
    # avoid circular import
    from pfb.operators.hessian import _hessian_slice, hess_direct_slice

    # expects a list
    if not isinstance(ds_name, list):
        ds_name = [ds_name]

    drop_vars = ['PSF', 'PSFHAT']
    ds = xds_from_list(ds_name, nthreads=nthreads,
                       drop_vars=drop_vars)[0]
    beam = mask * ds.BEAM.values
    if zero_model_outside_mask:
        if model_name not in ds:
            raise RuntimeError(f"Asked to zero model outside mask but {model_name} not in dds")
        model = getattr(ds, model_name).values
        model = np.where(mask > 0, model, 0.0)
        print("Zeroing model outside mask")
        resid = ds.DIRTY.values - _hessian_slice(
                                        model,
                                        uvw=ds.UVW.values,
                                        weight=ds.WEIGHT.values,
                                        vis_mask=ds.MASK.values,
                                        freq=ds.FREQ.values,
                                        beam=ds.BEAM.values,
                                        cell=ds.cell_rad,
                                        x0=ds.x0,
                                        y0=ds.y0,
                                        do_wgridding=do_wgridding,
                                        epsilon=epsilon,
                                        double_accum=double_accum,
                                        nthreads=nthreads)
        j = resid * beam
    else:
        if model_name in ds:
            model = getattr(ds, model_name).values
        else:
            model = np.zeros(mask.shape, dtype=float)

        if residual_name in ds:
            j = getattr(ds, residual_name).values * beam
            ds = ds.drop_vars(residual_name)
        else:
            j = ds.DIRTY.values * beam

    nx, ny = j.shape
    wsum = ds.wsum
    j /= wsum
    precond = None
    if 'UPDATE' in ds:
        x0 = ds.UPDATE.values * mask
    else:
        x0 = np.zeros_like(j)

    hess = partial(_hessian_slice,
                   uvw=ds.UVW.values,
                   weight=ds.WEIGHT.values,
                   vis_mask=ds.MASK.values,
                   freq=ds.FREQ.values,
                   beam=beam,
                   cell=ds.cell_rad,
                   x0=ds.x0,
                   y0=ds.y0,
                   flip_u=ds.flip_u,
                   flip_v=ds.flip_v,
                   flip_w=ds.flip_w,
                   do_wgridding=do_wgridding,
                   epsilon=epsilon,
                   double_accum=double_accum,
                   nthreads=nthreads,
                   eta=eta,
                   wsum=wsum)

    x = pcg(hess,
            j,
            x0=x0.copy(),
            M=precond,
            tol=tol,
            maxit=maxit,
            minit=1,
            verbosity=verbosity,
            report_freq=report_freq,
            backtrack=False,
            return_resid=False)

    model += x

    resid = ds.DIRTY.values - _hessian_slice(
                                        model,
                                        uvw=ds.UVW.values,
                                        weight=ds.WEIGHT.values,
                                        vis_mask=ds.MASK.values,
                                        freq=ds.FREQ.values,
                                        beam=ds.BEAM.values,
                                        cell=ds.cell_rad,
                                        x0=ds.x0,
                                        y0=ds.y0,
                                        do_wgridding=do_wgridding,
                                        epsilon=epsilon,
                                        double_accum=double_accum,
                                        nthreads=nthreads)

    ds = ds.assign(**{
        'MODEL_MOPPED': (('x','y'), model),
        'RESIDUAL_MOPPED': (('x','y'), resid),
        'UPDATE': (('x','y'), x),
        'X0': (('x','y'), x0),
    })

    ds.to_zarr(ds_name[0], mode='a')

    return resid, int(ds.bandid)





import numpy as np
from pyrap.measures import measures
from pyrap.quanta import quantity
from astropy.time import Time
from astropy import units
from astropy.coordinates import SkyCoord
from astropy.coordinates import solar_system_ephemeris, EarthLocation, AltAz
from astropy.coordinates import get_body_barycentric, get_body
from scipy.constants import c as lightspeed
from pfb.operators.gridder import wgridder_conventions
from africanus.coordinates import radec_to_lmn


# Based on
# https://github.com/tart-telescope/tart2ms/blob/master/tart2ms/fixvis.py
def synthesize_uvw(station_ECEF, time, ant1, ant2,
                   phase_ref,
                   stopctr_units=["rad", "rad"], stopctr_epoch="j2000",
                   time_TZ="UTC", time_unit="s",
                   posframe="ITRF", posunits=["m", "m", "m"]):
    """
    Synthesizes new UVW coordinates based on time according to
    NRAO CASA convention (same as in fixvis). 

    Note - this should work with missing rows as long as 
    
    inputs:
        station_ECEF: ITRF station coordinates read from MS::ANTENNA
        time: time column, preferably time centroid
        ant1: ANTENNA_1 index
        ant2: ANTENNA_2 index
        phase_ref: phase reference centre in radians
    
    returns uvw coordinates w.r.t. phase_ref
    """
    assert time.size == ant1.size
    assert ant1.size == ant2.size
    unique_time = np.unique(time)

    # ant to index mapping
    uants, iants = np.unique(np.concatenate((ant1, ant2)), return_index=True)
    # np.unique sorts the outputs, we need to unsort them
    # for the ant to index mapping to remain consistent
    isort = np.argsort(iants)
    uants = uants[isort]
    ants_dct = {uant: i for i, uant in enumerate(uants)}
    nrow = ant1.size
    uvw_new = np.zeros((nrow, 3))

    dm = measures()
    epoch = dm.epoch(time_TZ, quantity(unique_time[0], time_unit))
    refdir = dm.direction(stopctr_epoch,
                          quantity(phase_ref[0], stopctr_units[0]),
                          quantity(phase_ref[1], stopctr_units[1]))
    obs = dm.position(posframe,
                      quantity(station_ECEF[0, 0], posunits[0]),
                      quantity(station_ECEF[0, 1], posunits[1]),
                      quantity(station_ECEF[0, 2], posunits[2]))

    #setup local horizon coordinate frame with antenna 0 as reference position
    dm.do_frame(obs)
    dm.do_frame(refdir)
    dm.do_frame(epoch)
    for t in unique_time:
        epoch = dm.epoch(time_TZ, quantity(t, "s"))
        dm.do_frame(epoch)

        station_uv = np.zeros_like(station_ECEF)
        for iapos, apos in enumerate(station_ECEF):
            compuvw = dm.to_uvw(dm.baseline(posframe,
                                            quantity([apos[0], station_ECEF[0, 0]], posunits[0]),
                                            quantity([apos[1], station_ECEF[0, 1]], posunits[1]),
                                            quantity([apos[2], station_ECEF[0, 2]], posunits[2])))
            station_uv[iapos] = compuvw["xyz"].get_value()[0:3]
        
        rows = np.where(time == t)[0]

        for row in rows:
            a1 = ant1[row]
            a2 = ant2[row]
            bla1 = ants_dct[a1]
            bla2 = ants_dct[a2]
            uvw_new[row] = station_uv[bla1] - station_uv[bla2]

    return uvw_new


def rephase(vis, uvw, freq, radec_new, radec_ref, phasesign=-1):
    '''
    vis         - (nrow, nchan, ncorr) visibilities to rephase
    uvw         - (nrow, 3) visibility space coordinates
    freq        - (nchan) frequencies in Hz
    radec_new   - new phase center in radians
    radec_ref   - old phase center in radians
    phasesign   - opposite phase sign to im -> vis direction 
    '''
    ra = radec_new[0]
    dec = radec_new[1]
    ra0 = radec_ref[0]
    dec0 = radec_ref[1]
    dra = (ra - ra0)
    cos_dec = np.cos(dec)
    sin_dec = np.sin(dec)
    sin_dra = np.sin(dra)
    cos_dra = np.cos(dra)
    cos_dec0 = np.cos(dec0)
    sin_dec0 = np.sin(dec0)
    ll = cos_dec * sin_dra
    mm = sin_dec * cos_dec0 - cos_dec * sin_dec0 * cos_dra
    nn = sin_dec * sin_dec0 + cos_dec * cos_dec0 * cos_dra - 1.0

    nrow, _, _ = vis.shape
    uvw_freq = np.zeros((nrow, freq.size, 3))
    uvw_freq[:, :, 0] = uvw[:, 0:1] * freq[None, :]/lightspeed
    uvw_freq[:, :, 1] = uvw[:, 1:2] * freq[None, :]/lightspeed
    uvw_freq[:, :, 2] = uvw[:, 2:] * freq[None, :]/lightspeed

    # rephasing conventions should match wgridder
    flip_u, flip_v, flip_w, _, _ = wgridder_conventions(0, 0)
    usign = -1 if flip_u else 1
    vsign = -1 if flip_v else 1
    wsign = -1 if flip_w else 1
    # ll *= usign
    # mm *= vsign
    x = np.exp(phasesign * 2.0j * np.pi * (uvw_freq[:, :, 0] * ll * usign +
                                           uvw_freq[:, :, 1] * mm * vsign -
                                           uvw_freq[:, :, 2] * nn * wsign))
    
    return vis[:, :, :] * x[:, :, None]


def format_coords(ra0,dec0):
    c = SkyCoord(ra0*units.deg,dec0*units.deg,frame='fk5')
    hms = str(c.ra.to_string(units.hour))
    dms = str(c.dec)
    return hms,dms

# Pillaged from
# https://github.com/ratt-ru/solarkat/blob/main/solarkat-pipeline/find_sun_stimela.py
# obs_lat and obs_lon hardcoded for MeerKAT
def get_coordinates(obs_time,
                    obs_lat=-30.71323598930457,
                    obs_lon=21.443001467965008,
                    target='Sun'):
    '''
    Give location of object given telescope location (defaults to MeerKAT)
    and time of observation.

    Inputs:

    obs_time - should be weighted mean of TIME from MS
    obs_lat  - telescope lattitude in degrees
    obs_lon  - telescope longitude in degrees
    '''
    loc = EarthLocation.from_geodetic(obs_lat,obs_lon) #,obs_height,ellipsoid)
    t = Time(obs_time/86400.0,format='mjd')  # factor converts Jsecs to Jdays 24 * 60**2
    with solar_system_ephemeris.set('builtin'):
        sun = get_body(target, t, loc)
        sun_ra = sun.ra.value
        sun_dec = sun.dec.value
    # sun_hms=format_coords(sun_ra,sun_dec)
    # print(sun_hms[0],sun_hms[1])
    return np.deg2rad(sun_ra), np.deg2rad(sun_dec)


def create_cross_product_matrix(k):
    """
    Create the cross-product matrix (skew-symmetric matrix) for vector k.
    
    For a vector k = [kx, ky, kz], the cross-product matrix K is:
    K * v = k × v for any vector v
    
    Parameters:
    -----------
    k : array (3,)
        Vector to create cross-product matrix from
    
    Returns:
    --------
    K : array (3, 3)
        Cross-product matrix
    """
    return np.array([[0, -k[2], k[1]],
                     [k[2], 0, -k[0]],
                     [-k[1], k[0], 0]])


def create_rotation_matrix_rodrigues(s0, s1):
    """
    Create rotation matrix using Rodrigues' formula to transform UVW coordinates 
    from old to new phase center.
    
    Uses the hybrid approach: Rodrigues' formula to compute the matrix once,
    then apply matrix multiplication for efficiency with multiple baselines.
    
    Parameters:
    -----------
    ra0, dec0 : float
        Original phase center (radians)
    ra1, dec1 : float
        New phase center (radians)
    
    Returns:
    --------
    R : ndarray (3, 3)
        Rotation matrix
    """
    # Calculate rotation axis (perpendicular to both directions)
    k = np.cross(s0, s1)
    k_norm = np.linalg.norm(k)
    
    # Handle special case: phase centers are identical or opposite
    if k_norm < 1e-10:
        if np.dot(s0, s1) > 0:
            # Same direction - identity matrix
            return np.eye(3)
        else:
            # Opposite directions - 180 degree rotation
            # Need to find any perpendicular axis
            # Use the axis most perpendicular to s0
            if abs(s0[0]) < 0.9:
                k = np.cross(s0, np.array([1, 0, 0]))
            else:
                k = np.cross(s0, np.array([0, 1, 0]))
            k = k / np.linalg.norm(k)
            # 180 degree rotation matrix using Rodrigues
            K = create_cross_product_matrix(k)
            return np.eye(3) + 2.0 * K @ K
    
    # Normalize rotation axis
    k = k / k_norm
    
    # Calculate rotation angle
    cos_theta = np.dot(s0, s1)
    # Clamp to avoid numerical issues with arccos
    cos_theta = np.clip(cos_theta, -1.0, 1.0)
    sin_theta = k_norm  # |s0 × s1| = sin(θ) for unit vectors
    
    # Build rotation matrix using Rodrigues' formula:
    # R = I + sin(θ) * K + (1 - cos(θ)) * K²
    # where K is the cross-product matrix of k
    
    K = create_cross_product_matrix(k)
    K2 = K @ K
    
    R = np.eye(3) + sin_theta * K + (1 - cos_theta) * K2
    
    return R

def change_phase_dir(vis, uvw, freq, radec_new, radec_ref, phasesign=-1):
    # Direction cosines for new phase center
    l, m, n = radec_to_lmn(radec_new[None, :], radec_ref)[0]
    s1 = np.array([l, m, n])

    # Direction cosines for original phase center
    s0 = np.array([0, 0, 1])

    # Difference in direction cosines
    dl = s1[0]
    dm = -s1[1]
    dn = s1[2] - 1
    
    # Phase shift
    nrow, nchan, _ = vis.shape
    phase = np.zeros((nrow, nchan), dtype=uvw.dtype)
    flip_u, flip_v, flip_w = wgridder_conventions(0,0)[0:3]
    usign = -1 if flip_u else 1
    phase += usign * dl * uvw[:, 0:1] * freq[None, :]/lightspeed
    vsign = -1 if flip_v else 1
    phase += vsign * dm * uvw[:, 1:2] * freq[None, :]/lightspeed
    wsign = -1 if flip_w else 1
    phase -= wsign * dn * uvw[:, 2:3] * freq[None, :]/lightspeed
    vis *= np.exp(phasesign*2j*np.pi*phase)[:, :, None]

    # # get rotation matrix
    # R = create_rotation_matrix_rodrigues(s0, s1)

    # # rotate uvw's (einsum is slower)
    # uvw_new = (R @ uvw.T).T

    return vis, uvw
    

def uvw_rotate(uvw, ra0, dec0, ra, dec):
    """
        Compute the following 3x3 coordinate transformation matrix:
        Z_rot(facet_new_rotation) * \\
        T(new_phase_centre_ra,new_phase_centre_dec) * \\
        transpose(T(old_phase_centre_ra,
                    old_phase_centre_dec)) * \\
        transpose(Z_rot(facet_old_rotation))
        where:
                            |      cRA             -sRA            0       |
        T (RA,D) =          |      -sDsRA          -sDcRA          cD      |
                            |      cDsRA           cDcRA           sD      |
        This is the similar to the one in Thompson, A. R.; Moran, J. M.;
        and Swenson, G. W., Jr. Interferometry and Synthesis
        in Radio Astronomy, New York: Wiley, ch. 4, but in a
        lefthanded system.
        We're not transforming between a coordinate system with w pointing
        towards the pole and one with w pointing towards the reference
        centre here, so the last rotation matrix is ignored!
        This transformation will let the image be tangent to the celestial
        sphere at the new delay centre
    """
    d_ra = ra - ra0
    c_d_ra = np.cos(d_ra)
    s_d_ra = np.sin(d_ra)
    c_new_dec = np.cos(dec)
    c_old_dec = np.cos(dec0)
    s_new_dec = np.sin(dec)
    s_old_dec = np.sin(dec0)
    mat_11 = c_d_ra
    mat_12 = s_old_dec * s_d_ra
    mat_13 = -c_old_dec * s_d_ra
    mat_21 = -s_new_dec * s_d_ra
    mat_22 = s_new_dec * s_old_dec * c_d_ra + c_new_dec * c_old_dec
    mat_23 = -c_old_dec * s_new_dec * c_d_ra + c_new_dec * s_old_dec
    mat_31 = c_new_dec * s_d_ra
    mat_32 = -c_new_dec * s_old_dec * c_d_ra + s_new_dec * c_old_dec
    mat_33 = c_new_dec * c_old_dec * c_d_ra + s_new_dec * s_old_dec
    uvw[0] = mat_11 * uvw[0] + mat_12 * uvw[1] + mat_13 * uvw[3]
    uvw[1] = mat_21 * uvw[0] + mat_22 * uvw[1] + mat_23 * uvw[3]
    uvw[2] = mat_31 * uvw[0] + mat_32 * uvw[1] + mat_33 * uvw[3]
    return uvw


import numpy as np
from numba import njit
import dask.array as da
from xarray import Dataset
from operator import getitem
from pfb.utils.beam import interp_beam

def single_corr(ds=None,
                jones=None,
                opts=None,
                freq=None,
                chan_width=None,
                bandid=None,
                cell_rad=None,
                utimes=None,
                tbin_idx=None,
                tbin_counts=None,
                timeid=None,
                radec=None):

    if opts.precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif opts.precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128

    data = getattr(ds, opts.data_column).data
    nrow, nchan, _ = data.shape

    ant1 = ds.ANTENNA1.data
    ant2 = ds.ANTENNA2.data

    # MS may contain auto-correlations
    if 'FLAG_ROW' in ds:
        frow = ds.FLAG_ROW.data | (ant1 == ant2)
    else:
        frow = (ant1 == ant2)

    if opts.weight_column is not None:
        weight = getattr(ds, opts.weight_column).data
    else:
        weight = da.ones_like(data, dtype=real_type)

    if data.dtype != complex_type:
        data = data.astype(complex_type)

    if weight.dtype != real_type:
        weight = weight.astype(real_type)

    if jones is None:
        # TODO - remove unnecessary jones multiplies
        ntime = tbin_idx.size
        nant = da.maximum(ant1.max(), ant2.max()).compute() + 1
        jones = da.ones((ntime, nchan, nant, 1, 2),
                        chunks=(tbin_idx.chunks[0][0], -1, -1, 1, 2),
                        dtype=complex_type)
    elif jones.dtype != complex_type:
        jones = jones.astype(complex_type)

    # qcal has chan and ant axes reversed compared to pfb implementation
    jones = da.swapaxes(jones, 1, 2)

    vis, wgt = weight_data(data, weight, jones, tbin_idx, tbin_counts,
                           ant1, ant2)

    if opts.flag_column is not None:
        flag = getattr(ds, opts.flag_column).data
        flag = da.any(flag, axis=2)
        flag = da.logical_or(flag, frow[:, None])
    else:
        flag = da.broadcast_to(frow[:, None], (nrow, nchan))

    mask = ~flag
    uvw = ds.UVW.data

    data_vars = {'FREQ': (('chan',), freq)}

    wsum = da.sum(wgt[mask])
    data_vars['WSUM'] = (('scalar'), da.array((wsum,)))

    # wgt = wgt.rechunk({0:opts.row_out_chunk})
    data_vars['WEIGHT'] = (('row', 'chan'), wgt)

    # uvw = uvw.rechunk({0:opts.row_out_chunk})
    data_vars['UVW'] = (('row', 'uvw'), uvw)

    # vis = vis.rechunk({0:opts.row_out_chunk})
    data_vars['VIS'] = (('row', 'chan'), vis)

    # MASK = ~FLAG.astype(np.uint8) for wgridder convention
    # mask = mask.rechunk({0:opts.row_out_chunk})
    data_vars['MASK'] = (('row', 'chan'), mask.astype(np.uint8))

    # if opts.weight:
    #     wgt = da.where(mask, wgt, 0.0)
    #     # TODO - BDA over frequency
    #     if opts.bda_decorr < 1:
    #         raise NotImplementedError("BDA not working yet")
    #         from africanus.averaging.dask import bda

    #         w_avs = []
    #         uvw = uvw.compute()
    #         t = time.compute()
    #         a1 = ant1.compute()
    #         a2 = ant2.compute()
    #         intv = interval.compute()
    #         fr = frow.compute()[:, None, None]

    #         res = bda(ds.TIME.data,
    #                   ds.INTERVAL.data,
    #                   ant1, ant2,
    #                   uvw=uvw,
    #                   flag=f[:, :, None],
    #                   weight_spectrum=wgt[:, :, None],
    #                   chan_freq=freq,
    #                   chan_width=chan_width,
    #                   decorrelation=0.95,
    #                   min_nchan=freq.size)

    #         uvw = res.uvw.reshape(-1, nchan, 3)[:, 0, :]
    #         wgt = res.weight_spectrum.reshape(-1, nchan).squeeze()

    #         uvw = uvw.rechunk({0:opts.row_out_chunk})
    #         data_vars['UVW'] = (('row', 'uvw'), uvw)

    #     wgt = wgt.rechunk({0:opts.row_out_chunk})
    #     data_vars['WEIGHT'] = (('row', 'chan'), wgt)


    # TODO - interpolate beam in time and freq
    npix = int(np.deg2rad(opts.max_field_of_view)/cell_rad)
    freq_out = np.mean(freq)
    beam = interp_beam(freq_out/1e6, npix, npix, np.rad2deg(cell_rad), opts.beam_model)

    data_vars['BEAM'] = (('scalar'), beam)

    attrs = {
        'ra' : radec[0],
        'dec': radec[1],
        'fieldid': ds.FIELD_ID,
        'ddid': ds.DATA_DESC_ID,
        'scanid': ds.SCAN_NUMBER,
        'bandid': int(bandid),
        'freq_out': freq_out
    }

    out_ds = Dataset(data_vars, attrs=attrs)

    return out_ds


def weight_data(data, weight, jones, tbin_idx, tbin_counts,
                ant1, ant2):
    # data are not necessarily 2x2 so we need separate labels
    # for jones correlations and data/weight correlations
    if jones.ndim == 5:
        jout = 'rafdx'
    elif jones.ndim == 6:
        jout = 'rafdxx'
        # TODO - how do we know if we should return
        # jones[0][0][0] or jones[0][0][0][0] in function wrapper?
        # Not required with delayed
        raise NotImplementedError("Not yet implemented")
    res = da.blockwise(_weight_data, 'rf',
                       data, 'rfc',
                       weight, 'rfc',
                       jones, jout,
                       tbin_idx, 'r',
                       tbin_counts, 'r',
                       ant1, 'r',
                       ant2, 'r',
                       align_arrays=False,
                       meta=np.empty((0, 0), dtype=object))

    vis = da.blockwise(getitem, 'rf', res, 'rf', 0, None, dtype=data.dtype)
    wgt = da.blockwise(getitem, 'rf', res, 'rf', 1, None, dtype=weight.dtype)


    return vis, wgt

def _weight_data(data, weight, jones, tbin_idx, tbin_counts,
                      ant1, ant2):
    return _weight_data_impl(data[0], weight[0], jones[0][0][0],
                             tbin_idx, tbin_counts, ant1, ant2)

@njit(nogil=True, cache=True)
def _weight_data_impl(data, weight, jones, tbin_idx, tbin_counts,
                      ant1, ant2):
    # for dask arrays we need to adjust the chunks to
    # start counting from zero
    tbin_idx -= tbin_idx.min()
    nt = np.shape(tbin_idx)[0]
    nrow, nchan, ncorr = data.shape
    vis = np.zeros((nrow, nchan), dtype=data.dtype)
    wgt = np.zeros((nrow, nchan), dtype=data.real.dtype)

    for t in range(nt):
        for row in range(tbin_idx[t],
                            tbin_idx[t] + tbin_counts[t]):
            p = int(ant1[row])
            q = int(ant2[row])
            gp = jones[t, p, :, 0]
            gq = jones[t, q, :, 0]
            for chan in range(nchan):
                wval = wgt_func(gp[chan], gq[chan],
                                weight[row, chan])
                wgt[row, chan] = wval
                vis[row, chan] = vis_func(gp[chan], gq[chan],
                                            weight[row, chan],
                                            data[row, chan])  #/wval

    return vis, wgt


@njit(nogil=True, cache=True, inline='always')
def wgt_func(gp, gq, W):
    gp00 = gp[0]
    gq00 = gq[0]
    W0 = W[0]
    return np.real(W0*gp00*gq00*np.conjugate(gp00)*np.conjugate(gq00))

@njit(nogil=True, cache=True, inline='always')
def vis_func(gp, gq, W, V):
    gp00 = gp[0]
    gq00 = gq[0]
    W0 = W[0]
    v00 = V[0]
    return W0*gq00*v00*np.conjugate(gp00)



import numpy as np
import numexpr as ne
from functools import partial
from numba import njit, prange
import dask.array as da
from pfb.operators.psf import psf_convolve_cube, psf_convolve_fscube
from ducc0.misc import empty_noncritical
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('CLARK')


@njit(nogil=True, cache=True, parallel=True)
def subminor(A, psf, Ip, Iq, model, wsums, gamma=0.05, th=0.0, maxit=10000):
    """
    Run subminor loop in active set

    A       - active set (all pixels above subminor threshold)
    psf     - psf image
    Ip      - x indices of active set
    Iq      - y indices of active set
    model   - current model image
    gamma   - loop gain
    th      - threshold to clean down to
    maxit   - maximum number of iterations

    Pixels in A map to pixels in the image via

    p = Ip[pq]
    q = Iq[pq]

    where pq is the location of the maximum in A.
    """
    nband, nx_psf, ny_psf = psf.shape
    nxo2 = nx_psf//2
    nyo2 = ny_psf//2
    # _, nx, ny = model.shape
    Asearch = np.sum(A, axis=0)**2
    pq = Asearch.argmax()
    p = Ip[pq]
    q = Iq[pq]
    Amax = np.sqrt(Asearch[pq])
    fsel = wsums > 0
    if fsel.sum() == 0:
        raise ValueError("wsums are all zero")
    k = 0
    while Amax > th and k < maxit:
        xhat = A[:, pq]
        model[fsel, p, q] += gamma * xhat[fsel]/wsums[fsel]
        
        for b in prange(nband):
            for i in range(Ip.size):
                pp = nxo2 - (Ip[i] - p)
                qq = nyo2 - (Iq[i] - q)
                if (pp >= 0) & (pp < nx_psf) & (qq >= 0) & (qq < ny_psf):
                    A[b, i] -= gamma * xhat[b] * psf[b, pp, qq]/wsums[b]

        Asearch = np.sum(A, axis=0)**2
        pq = Asearch.argmax()
        p = Ip[pq]
        q = Iq[pq]
        Amax = np.sqrt(Asearch[pq])
        k += 1
    return model

def clark(ID,
          PSF,
          PSFHAT,
          wsums,
          mask,
          threshold=0,
          gamma=0.05,
          pf=0.05,
          maxit=50,
          subpf=0.5,
          submaxit=1000,
          report_freq=1,
          verbosity=1,
          nthreads=1):
    nband, nx, ny = ID.shape
    _, nx_psf, ny_psf = PSF.shape
    # we assume that the dirty image and PSF have been normalised by wsum
    # and that we get units of Jy/beam when we take the sum over the frequency
    # axis i.e. the MFS image is in units of Jy/beam
    wsum = wsums.sum()
    assert np.allclose(wsum, 1)
    model = np.zeros((nband, nx, ny), dtype=ID.dtype)
    IR = ID.copy()
    # pre-allocate arrays for doing FFT's
    xout = empty_noncritical(ID.shape, dtype='f8')
    xpad = empty_noncritical(PSF.shape, dtype='f8')
    xhat = empty_noncritical(PSFHAT.shape, dtype='c16')
    # square avoids abs of full array
    IRsearch = np.sum(IR, axis=0)**2 * mask
    pq = IRsearch.argmax()
    p = pq//ny
    q = pq - p*ny
    IRmax = np.sqrt(IRsearch[p, q])
    tol = np.maximum(pf * IRmax, threshold)
    k = 0
    stall_count = 0
    while IRmax > tol and k < maxit and stall_count < 5:
        # identify active set
        subth = subpf * IRmax
        Ip, Iq = np.where(IRsearch > subth**2)
        # run substep in active set
        model = subminor(IR[:, Ip, Iq], PSF, Ip, Iq, model, wsums,
                         gamma=gamma,
                         th=subth,
                         maxit=submaxit)

        # subtract from full image (as in major cycle)
        psf_convolve_cube(
                        xpad,
                        xhat,
                        xout,
                        PSFHAT,
                        ny_psf,
                        model,
                        nthreads=nthreads)
        IR = ID - xout
        IRsearch = np.sum(IR, axis=0)**2 * mask
        pq = IRsearch.argmax()
        p = pq//ny
        q = pq - p*ny
        IRmaxp = IRmax
        IRmax = np.sqrt(IRsearch[p, q])
        k += 1

        if np.abs(IRmaxp - IRmax) / np.abs(IRmaxp) < 1e-3:
            stall_count += stall_count

        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} max resid = {IRmax}")

    IRmfs = np.sum(IR, axis=0)
    rms = np.std(IRmfs[~np.any(model, axis=0)])

    if k >= maxit:
        if verbosity:
            log.info(f"Max iters reached. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 1
    elif stall_count >= 5:
        if verbosity:
            log.info(f"Stalled. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 1
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 0


@njit(nogil=True, cache=True, parallel=True)
def fssubminor(A, psf, Ip, Iq, model, wsums, gamma=0.05, th=0.0, maxit=10000):
    """
    Full Stokes subminor loop in active set

    A       - active set (all pixels above subminor threshold)
    psf     - psf image
    Ip      - x indices of active set
    Iq      - y indices of active set
    model   - current model image
    gamma   - loop gain
    th      - threshold to clean down to
    maxit   - maximum number of iterations

    Pixels in A map to pixels in the image via

    p = Ip[pq]
    q = Iq[pq]

    where pq is the location of the maximum in A.
    """
    nband, ncorr, nx_psf, ny_psf = psf.shape
    nxo2 = nx_psf//2
    nyo2 = ny_psf//2
    # _, nx, ny = model.shape
    # MFS image
    Amfs = np.sum(A, axis=0)
    # total pol image (always positive)
    Asearch = np.sum(Amfs**2, axis=0)
    pq = Asearch.argmax()
    p = Ip[pq]
    q = Iq[pq]
    Amax = np.sqrt(Asearch[pq])
    fsel = wsums > 0
    if fsel.sum() == 0:
        raise ValueError("wsums are all zero")
    k = 0
    while Amax > th and k < maxit:
        xhat = A[:, :, pq]
        model[:, :, p, q] += gamma * xhat[:, :]/wsums[:, :]
        
        for b in prange(nband):
            for c in range(ncorr):
                for i in range(Ip.size):
                    pp = nxo2 - (Ip[i] - p)
                    qq = nyo2 - (Iq[i] - q)
                    if (pp >= 0) & (pp < nx_psf) & (qq >= 0) & (qq < ny_psf):
                        A[b, c, i] -= gamma * xhat[b, c] * psf[b, c, pp, qq]/wsums[b, c]

        Amfs = np.sum(A, axis=0)
        Asearch = np.sum(Amfs**2, axis=0)
        pq = Asearch.argmax()
        p = Ip[pq]
        q = Iq[pq]
        Amax = np.sqrt(Asearch[pq])
        k += 1
    return model


def fsclark(ID,
            PSF,
            PSFHAT,
            wsums,
            mask,
            threshold=0,
            gamma=0.05,
            pf=0.05,
            maxit=50,
            subpf=0.5,
            submaxit=1000,
            report_freq=1,
            verbosity=1,
            nthreads=1):
    nband, ncorr, nx, ny = ID.shape
    _, _, nx_psf, ny_psf = PSF.shape
    # we assume that the dirty image and PSF have been normalised by wsum
    # and that we get units of Jy/beam when we take the sum over the frequency
    # axis i.e. the MFS image is in units of Jy/beam
    wsum = wsums.sum(axis=0)
    assert np.allclose(wsum, 1)
    model = np.zeros((nband, ncorr, nx, ny), dtype=ID.dtype)
    IR = ID.copy()
    # pre-allocate arrays for doing FFT's
    xout = empty_noncritical(ID.shape, dtype='f8')
    xpad = empty_noncritical(PSF.shape, dtype='f8')
    xhat = empty_noncritical(PSFHAT.shape, dtype='c16')
    # MFS image
    IRmfs = np.sum(IR, axis=0)
    # total pol image (always positive)
    # Do we technically need a weighted sum here?
    IRsearch = np.sum(IRmfs**2, axis=0) * mask
    pq = IRsearch.argmax()
    p = pq//ny
    q = pq - p*ny
    IRmax = np.sqrt(IRsearch[p, q])
    tol = np.maximum(pf * IRmax, threshold)
    k = 0
    stall_count = 0
    while IRmax > tol and k < maxit and stall_count < 5:
        # identify active set
        subth = subpf * IRmax
        Ip, Iq = np.where(IRsearch > subth**2)
        # run substep in active set
        model = fssubminor(IR[:, :, Ip, Iq], PSF, Ip, Iq, model, wsums,
                           gamma=gamma,
                           th=subth,
                           maxit=submaxit)

        # subtract from full image (as in major cycle)
        psf_convolve_fscube(
                        xpad,
                        xhat,
                        xout,
                        PSFHAT,
                        ny_psf,
                        model,
                        nthreads=nthreads)
        IR = ID - xout
        IRmfs = np.sum(IR, axis=0)
        IRsearch = np.sum(IRmfs**2, axis=0) * mask
        pq = IRsearch.argmax()
        p = pq//ny
        q = pq - p*ny
        IRmaxp = IRmax
        IRmax = np.sqrt(IRsearch[p, q])
        k += 1

        if np.abs(IRmaxp - IRmax) / np.abs(IRmaxp) < 1e-3:
            stall_count += stall_count

        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} max resid = {IRmax}")

    IRmfs = np.sum(IR, axis=0)
    rms = np.std(IRmfs, axis=(-2,-1)).max()

    if k >= maxit:
        if verbosity:
            log.info(f"Max iters reached. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 1
    elif stall_count >= 5:
        if verbosity:
            log.info(f"Stalled. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 1
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations. "
                     f"Max resid = {IRmax:.3e}, rms = {rms:.3e}")
        return model, 0


import numba


# from the answer given here
# https://stackoverflow.com/questions/67431966/how-to-avoid-huge-overhead-of-single-threaded-numpys-transpose
@numba.njit(nogil=True, cache=True, parallel=True, inline='always')
def copyT(mat, out):
    blockSize, tileSize = 256, 32  # To be tuned
    n, m = mat.shape
    for tmp in numba.prange((m+blockSize-1)//blockSize):
        i = tmp * blockSize
        for j in range(0, n, blockSize):
            tiMin, tiMax = i, min(i+blockSize, m)
            tjMin, tjMax = j, min(j+blockSize, n)
            for ti in range(tiMin, tiMax, tileSize):
                for tj in range(tjMin, tjMax, tileSize):
                    out[ti:ti+tileSize, tj:tj+tileSize] = mat[tj:tj+tileSize, ti:ti+tileSize].T


@numba.njit(nogil=True, cache=True)
def coeff_size(nsignal, nfilter):
    return (nsignal + nfilter - 1)//2


@numba.njit(nogil=True, cache=True)
def signal_size(ncoeff, nfilter):
    return 2*ncoeff - nfilter + 2


@numba.njit(nogil=True, cache=True, inline='always', fastmath=False, error_model='numpy')
def downsampling_convolution(input, output, filter, step):

    i = step - 1
    o = 0
    N = input.shape[0]
    F = filter.shape[0]


    # left boundary overhang
    while i < F and i < N:
        fsum = input.dtype.type(0)  # init to zero with correct type
        j = 0

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if input equal or wider than filter: N >= F)
    while i < N:
        fsum = input.dtype.type(0)
        j = 0

        while j < F:
            fsum += input[i - j] * filter[j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if filter is wider than input: F > N)
    while i < F:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # right boundary overhang
    while i < N + F - 1:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j < F:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1


@numba.njit(nogil=True, cache=True, inline='always', fastmath=False, error_model='numpy')
def upsampling_convolution_valid_sf(input, filter, output):

    N = input.shape[0]
    F = filter.shape[0]
    O = output.shape[0]

    # if F % 2:
    #     raise ValueError("even filter required for upsampling")

    # if N < (F // 2):
    #     raise ValueError("input.shape[0] < (filter.shape[0] // 2)")

    o = 0
    i = (F // 2) - 1
    stopping_criteria = F // 2

    while i < N and o < O:
        sum_even = input.dtype.type(0)
        sum_odd = input.dtype.type(0)
        j = 0
        j2 = 0

        while j < stopping_criteria:
            input_element = input[i - j]
            sum_even += filter[j2] * input_element
            sum_odd += filter[j2 + 1] * input_element

            j += 1
            j2 += 2

        output[o] += sum_even
        output[o + 1] += sum_odd

        i += 1
        o += 2


@numba.njit(nogil=True, cache=True, parallel=True)
def dwt2d_level(image, coeffs, cbuff, cbuffT, dec_lo, dec_hi):
    """
    Map image to coeffs for a single level

    image   - (nx, ny) signal
    coeffs  - (nay, nax) output coeffs
    cbuff   - (nax, nay) coeff buffer
    cbuffT  - (nay, nax) coeff buffer
    dec_lo  - (F) length low pass filter for decomposition
    dec_hi  - (F) length high pass filter for decomposition

    The dimension of the output is given by

    nax = (nx + F - 1)//2
    nay = (ny + F - 1)//2

    The output is transposed with respect to the input for
    performance reasons.

    This function is not meant to be called directly.
    """
    nx, ny = image.shape
    nay, nax = coeffs.shape

    midy = nay//2
    for i in numba.prange(nx):
        # approx
        downsampling_convolution(image[i, :], cbuff[i, 0:midy], dec_lo, 2)
        # detail
        downsampling_convolution(image[i, :], cbuff[i, midy:], dec_hi, 2)

    # prefer over repeatedly convolving the non-contiguous array
    copyT(cbuff, cbuffT)

    midx = nax//2
    for i in numba.prange(nay):
        # approx
        downsampling_convolution(cbuffT[i, 0:nx], coeffs[i, 0:midx], dec_lo, 2)
        # detail
        downsampling_convolution(cbuffT[i, 0:nx], coeffs[i, midx:], dec_hi, 2)

    # approx coeffs are in the top left corner of out
    # copy and transpose to ensure correct order and
    # that input is not overwritten at the next level
    return coeffs[0:midy, 0:midx].T.copy()


@numba.njit(nogil=True, cache=True, parallel=True)
def dwt2d(image, coeffs, cbuff, cbuffT, ix, iy,
          sx, sy, dec_lo, dec_hi, nlevel):
    '''
    Multi-level 2D image to coeffs transform

    image   - (nx, ny) signal
    coeffs  - (nay, nax) output coeffs
    cbuff   - (nax, nay) coeff buffer
    cbuffT  - (nay, nax) coeff buffer
    ix,iy   - (nlevel) dicts containing start and stop indices for packing coeffs
    sx,sy   - (nlevel) tuples containing coeff sizes at each level
    spx,spy - (nlevel) tuples containing signal sizes at each level
    dec_lo  - (F) length low pass filter for decomposition
    dec_hi  - (F) length high pass filter for decomposition
    nlevel  - the number of decomposition levels

    The dimension of the output is given by

    nax = (nx + F - 1)//2
    nay = (ny + F - 1)//2

    The output is transposed with respect to the input for
    performance reasons.

    '''
    nx, ny = image.shape
    approx = image
    for i in range(nlevel):
        _, highx = ix[i]
        lowx = highx - 2*sx[i]
        _, highy = iy[i]
        lowy = highy - 2*sy[i]
        # detail coeffs go directly into output
        # returned approx coeffs are safe to overwrite since copied
        approx = dwt2d_level(approx,
                             coeffs[lowy:highy, lowx:highx],
                             cbuff[lowx:highx, lowy:highy],
                             cbuffT[lowy:highy, lowx:highx],
                             dec_lo, dec_hi)


@numba.njit(nogil=True, cache=True, parallel=True)
def idwt2d_level(coeffs, image, cbuff, cbuffT, rec_lo, rec_hi):
    '''
    Map coeffs to image for a single level

    coeffs  - (nay, nax) coeffs
    image   - (nx, ny) output signal
    cbuff   - (nax, nay) coeff buffer
    cbuffT  - (nay, nax) coeff buffer
    rec_lo  - (F) length low pass filter for reconstruction
    rec_hi  - (F) length high pass filter for reconstruction

    The dimension of the output signal is give by

    nx = 2*nax - F + 2
    ny = 2*nay - F + 2

    The output is transposed with respect to the input for
    performance reasons.

    This function is not meant to be called directly.

    '''
    nay, nax = coeffs.shape
    nx, ny = image.shape

    # zero since accumulated into
    cbuffT[...] = 0.0
    image[...] = 0.0

    midx = nax//2
    for i in numba.prange(nay):
        upsampling_convolution_valid_sf(coeffs[i, 0:midx], rec_lo, cbuffT[i, :])
        upsampling_convolution_valid_sf(coeffs[i, midx:], rec_hi, cbuffT[i, :])


    copyT(cbuffT, cbuff)

    midy = nay//2
    for i in numba.prange(nx):
        upsampling_convolution_valid_sf(cbuff[i, 0:midy], rec_lo, image[i, :])
        upsampling_convolution_valid_sf(cbuff[i, midy:], rec_hi, image[i, :])


@numba.njit(nogil=True, cache=True, parallel=True)
def idwt2d(coeffs, image, alpha, cbuff, cbuffT, ix, iy,
           sx, sy, spx, spy, rec_lo, rec_hi, nlevel):
    '''
    Multi-level 2D coeffs to image transform

    coeffs  - (nay, nax) output coeffs
    image   - (nx, ny) signal
    alpha   - (nay, nax) coeff buffer
    cbuff   - (nax, nay) coeff buffer
    cbuffT  - (nay, nax) coeff buffer
    ix,iy   - (nlevel) dicts containing start and stop indices for packing coeffs
    sx,sy   - (nlevel) tuples containing coeff sizes at each level
    spx,spy - (nlevel) tuples containing signal sizes at each level
    dec_lo  - (F) length low pass filter for decomposition
    dec_hi  - (F) length high pass filter for decomposition
    nlevel  - the number of decomposition levels

    The dimension of the output is given by

    nx = 2*nax - F + 2
    ny = 2*nay - F + 2

    The output is transposed with respect to the input for
    performance reasons.

    '''
    nx, ny = image.shape
    alpha[...] = coeffs  # to avoid overwriting coeffs
    for i in range(nlevel-1, -1, -1):
        # coeff size at current level
        nax = sx[i]
        nay = sy[i]

        # slice indices for alpha
        _, highx = ix[i]
        lowx = highx - 2*nax
        _, highy = iy[i]
        lowy = highy - 2*nay

        # signal size at current level
        nxo = spx[i]
        nyo = spy[i]

        # previous output is the top left corner of the next input
        # note the difference in indexing (nax,nay instead of nxo,nyo)
        # that stems from the redundant coefficient
        if i < nlevel - 1:
            copyT(image[0:nax, 0:nay], alpha[lowy:lowy+nay, lowx:lowx+nax])

        # need a buffer the size of the coeffs at next level
        idwt2d_level(alpha[lowy:highy, lowx:highx],
                     image[0:nxo, 0:nyo],
                     cbuff[0:2*nax, 0:2*nay],
                     cbuffT[0:2*nay, 0:2*nax],
                     rec_lo, rec_hi)



# flake8: noqa
import click
import time
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('DEGRID')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.degrid)
def degrid(**kw):
    '''
    Predict model visibilities to measurement sets.
    The default behaviour is to read the frequency mapping from the dds and
    degrid one image per band.
    If channels-per-image is provided, the model is evaluated from the mds.
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    from daskms.fsspec_store import DaskMSStore
    msnames = []
    for ms in opts.ms:
        msstore = DaskMSStore(ms.rstrip('/'))
        mslist = msstore.fs.glob(ms.rstrip('/'))
        try:
            assert len(mslist) > 0
            msnames.append(*list(map(msstore.fs.unstrip_protocol, mslist)))
        except:
            log.error_and_raise(f"No MS at {ms}", ValueError)
    opts.ms = msnames

    basename = opts.output_filename

    if opts.mds is None:
        mds_store = DaskMSStore(f'{basename}_{opts.suffix}_model.mds')
    else:
        mds_store = DaskMSStore(opts.mds)
        try:
            assert mds_store.exists()
        except Exception as e:
            log.error_and_raise(f"No mds at {opts.mds}", ValueError)
    opts.mds = mds_store.url

    dds_store = DaskMSStore(f'{basename}_{opts.suffix}.dds')
    if opts.channels_per_image is None and not mds_store.exists():
        try:
            assert dds_store.exists()
        except Exception as e:
            log.error_and_raise(f"There must be a dds at {dds_store.url}. "
                                "Specify mds and channels-per-image to degrid from mds.",
                                ValueError)
    opts.dds = dds_store.url
    OmegaConf.set_struct(opts, True)

    remprod = opts.product.upper().strip('IQUV')
    if len(remprod):
        log.error_and_raise(f"Product {remprod} not yet supported",
                            NotImplementedError)

    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/degrid_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    # we need the collections
    from pfb import set_client
    client = set_client(opts.nworkers, log, None, client_log_level=opts.log_level)

    ti = time.time()
    _degrid(**opts)

    log.info(f"All done after {time.time() - ti}s.")

    try:
        client.close()
    except Exception as e:
        pass

def _degrid(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    from pfb.utils.misc import construct_mappings
    import dask
    from dask.distributed import wait, get_client
    from dask.graph_manipulation import clone
    from daskms import xds_from_storage_ms as xds_from_ms
    from daskms import xds_to_storage_table as xds_to_table
    from daskms.fsspec_store import DaskMSStore
    import dask.array as da
    from pfb.operators.gridder import comps2vis
    from pfb.utils.fits import set_wcs
    from regions import Regions
    from pfb.utils.naming import xds_from_url
    import xarray as xr
    import sympy as sm
    from sympy.utilities.lambdify import lambdify
    from sympy.parsing.sympy_parser import parse_expr
    from africanus.model.coherency.dask import convert
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)

    client = get_client()

    dds_store = DaskMSStore(opts.dds)
    mds_store = DaskMSStore(opts.mds)

    if opts.channels_per_image is None:
        if dds_store.exists():
            dds, dds_list = xds_from_url(dds_store.url)
            cpi = 0
            for ds in dds:
                cpi = np.maximum(ds.chan.size, cpi)
        else:
            log.error_and_raise("You must supply channels per image in the "
                                "absence of a dds", ValueError)
    else:
        cpi = opts.channels_per_image

    if opts.freq_range is not None and len(opts.freq_range):
        fmin, fmax = opts.freq_range.strip(' ').split(':')
        if len(fmin) > 0:
            freq_min = float(fmin)
        else:
            freq_min = -np.inf
        if len(fmax) > 0:
            freq_max = float(fmax)
        else:
            freq_max = np.inf
    else:
        freq_min = -np.inf
        freq_max = np.inf


    group_by = ['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER']

    log.info('Constructing mapping')
    row_mapping, freq_mapping, time_mapping, \
        freqs, utimes, ms_chunks, gains, radecs, \
        chan_widths, uv_max, antpos, poltype = \
            construct_mappings(opts.ms,
                               None,
                               ipi=opts.integrations_per_image,
                               cpi=cpi,
                               freq_min=freq_min,
                               freq_max=freq_max)
                            #    FIELD_IDs=opts.fields,
                            #    DDIDs=opts.ddids,
                            #    SCANs=opts.scans)

    mds = xr.open_zarr(opts.mds)
    foo = client.scatter(mds, broadcast=True)
    wait(foo)

    # grid spec
    cell_rad = mds.cell_rad_x
    cell_deg = np.rad2deg(cell_rad)
    nx = mds.npix_x
    ny = mds.npix_y
    x0 = mds.center_x
    y0 = mds.center_y
    radec = (mds.ra, mds.dec)

    # model func
    params = sm.symbols(('t','f'))
    params += sm.symbols(tuple(mds.params.values))
    symexpr = parse_expr(mds.parametrisation)
    modelf = lambdify(params, symexpr)
    texpr = parse_expr(mds.texpr)
    tfunc = lambdify(params[0], texpr)
    fexpr = parse_expr(mds.fexpr)
    ffunc = lambdify(params[1], fexpr)

    # load region file if given
    masks = []
    if opts.region_file is not None:
        rfile = Regions.read(opts.region_file)  # should detect format
        # get wcs for model
        wcs = set_wcs(np.rad2deg(mds.cell_rad_x),
                      np.rad2deg(mds.cell_rad_y),
                      mds.npix_x,
                      mds.npix_y,
                      (mds.ra, mds.dec),
                      mds.freqs.values,
                      header=False)
        wcs = wcs.dropaxis(-1)
        wcs = wcs.dropaxis(-1)

        mask = np.zeros((nx, ny), dtype=np.float64)
        # get a mask for each region
        for region in rfile:
            pixel_region = region.to_pixel(wcs)
            # why the transpose?
            region_mask = pixel_region.to_mask().to_image((ny, nx))
            region_mask = region_mask.T
            mask += region_mask
            masks.append(region_mask)
        if (mask > 1).any():
            log.error_and_raise("Overlapping regions are not supported",
                                ValueError)
        remainder = 1 - mask
        # place DI component first
        masks = [remainder] + masks
    else:
        masks = [np.ones((nx, ny), dtype=np.float64)]

    input_schema = sorted(opts.product.upper())
    if poltype=='linear':
        output_schema = ['XX', 'XY', 'YX', 'YY']
    else:
        output_schema = ['RR', 'RL', 'LR', 'LL']

    writes = []
    for ms in opts.ms:
        xds = xds_from_ms(ms,
                          chunks=ms_chunks[ms],
                          group_cols=group_by)


        for i, mask in enumerate(masks):
            out_data = []
            columns = []
            for k, ds in enumerate(xds):
                fid = ds.FIELD_ID
                ddid = ds.DATA_DESC_ID
                scanid = ds.SCAN_NUMBER
                if (opts.fields is not None) and (fid not in opts.fields):
                    continue
                if (opts.ddids is not None) and (ddid not in opts.ddids):
                    continue
                if (opts.scans is not None) and (scanid not in opts.scans):
                    continue
                idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"

                if i == 0:
                    column_name = opts.model_column
                else:
                    column_name = f'{opts.model_column}{i}'
                columns.append(column_name)

                # time <-> row mapping
                utime = da.from_array(utimes[ms][idt],
                                      chunks=opts.integrations_per_image)
                tidx = da.from_array(time_mapping[ms][idt]['start_indices'],
                                     chunks=1)
                tcnts = da.from_array(time_mapping[ms][idt]['counts'],
                                      chunks=1)

                ridx = da.from_array(row_mapping[ms][idt]['start_indices'],
                                     chunks=opts.integrations_per_image)
                rcnts = da.from_array(row_mapping[ms][idt]['counts'],
                                      chunks=opts.integrations_per_image)

                # freq <-> band mapping (entire freq axis)
                freq = da.from_array(freqs[ms][idt],
                                     chunks=ms_chunks[ms][k]['chan'])
                fcnts = np.array(ms_chunks[ms][k]['chan'])
                fidx = np.concatenate((np.array([0]), np.cumsum(fcnts)))[0:-1]

                fidx = da.from_array(fidx,
                                     chunks=1)
                fcnts = da.from_array(fcnts,
                                      chunks=1)

                # number of chunks need to match in mapping and coord
                ntime_out = len(tidx.chunks[0])
                assert len(utime.chunks[0]) == ntime_out
                nfreq_out = len(fidx.chunks[0])
                assert len(freq.chunks[0]) == nfreq_out
                # and they need to match the number of row chunks
                uvw = clone(ds.UVW.data)
                assert len(uvw.chunks[0]) == len(tidx.chunks[0])

                ncorr = ds.corr.size

                vis = comps2vis(uvw,
                                utime,
                                freq,
                                ridx, rcnts,
                                tidx, tcnts,
                                fidx, fcnts,
                                mask,
                                mds,
                                modelf,
                                tfunc,
                                ffunc,
                                nthreads=opts.nthreads,
                                epsilon=opts.epsilon,
                                do_wgridding=opts.do_wgridding,
                                freq_min=freq_min,
                                freq_max=freq_max,
                                product=opts.product)

                # convert to single precision to write to MS
                vis = vis.astype(np.complex64)

                if ncorr==1:
                    out_schema = output_schema[0]
                elif ncorr==2:
                    out_schema = [output_schema[0], output_schema[-1]]
                else:
                    out_schema = output_schema

                vis = convert(vis, input_schema, out_schema, implicit_stokes=True)

                if opts.accumulate:
                    vis += getattr(ds, column_name).data

                out_ds = ds.assign(**{column_name:
                                    (("row", "chan", "corr"), vis)})
                out_data.append(out_ds)

            writes.append(xds_to_table(
                                    out_data, ms,
                                    columns=columns,
                                    rechunk=True))

    # optimize_graph can make things much worse
    log.info("Computing model visibilities")
    dask.compute(writes)  #, optimize_graph=False)




# flake8: noqa
import click
from omegaconf import OmegaConf
import time
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('KCLEAN')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.kclean)
def kclean(**kw):
    '''
    Modified single-scale clean.
    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        opts.nthreads = nthreads//2
        ncpu = ncpu//2

    OmegaConf.set_struct(opts, True)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/kclean_{timestamp}.log'
    pfb_logging.log_to_file(logname)

    pfb_logging.log_options_dict(log, opts)

    from pfb.utils.naming import xds_from_url, get_opts

    basename = f'{basedir}/{oname}'
    fits_oname = f'{opts.fits_output_folder}/{oname}'
    dds_name = f'{basename}_{opts.suffix}.dds'

    ti = time.time()
    if opts.product != 'I':
        _fskclean(**opts)
    else:
        _kclean(**opts)

    dds, dds_list = xds_from_url(dds_name)

    if opts.fits_mfs or opts.fits:
        from daskms.fsspec_store import DaskMSStore
        from pfb.utils.fits import dds2fits
        # get the psfpars for the mfs cube
        dds_store = DaskMSStore(dds_name)
        if '://' in dds_store.url:
            protocol = dds_store.url.split('://')[0]
        else:
            protocol = 'file'
        psfpars_mfs = get_opts(dds_store.url,
                               protocol,
                               name='psfparsn_mfs.pkl')
        log.info(f"Writing fits files to {fits_oname}_{opts.suffix}")
        dds2fits(dds_list,
                 'RESIDUAL',
                 f'{fits_oname}_{opts.suffix}',
                 norm_wsum=True,
                 do_mfs=opts.fits_mfs,
                 do_cube=opts.fits_cubes,
                 psfpars_mfs=psfpars_mfs)
        dds2fits(dds_list,
                 'MODEL',
                 f'{fits_oname}_{opts.suffix}',
                 norm_wsum=False,
                 do_mfs=opts.fits_mfs,
                 do_cube=opts.fits_cubes,
                 psfpars_mfs=psfpars_mfs)

    log.info(f"All done after {time.time() - ti}s")


def _kclean(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    from functools import partial
    import numpy as np
    import xarray as xr
    from pfb.utils.fits import set_wcs, save_fits, load_fits
    from pfb.deconv.clark import clark
    from pfb.utils.naming import xds_from_url
    from pfb.operators.hessian import hess_psf
    from pfb.operators.gridder import compute_residual
    from scipy import ndimage
    from pfb.utils.modelspec import fit_image_cube

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds, dds_list = xds_from_url(dds_name)

    if dds[0].corr.size > 1:
        log.error_and_raise("Joint polarisation deconvolution not yet supported for kclean algorithm",
                            NotImplementedError)

    nx, ny = dds[0].x.size, dds[0].y.size
    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf
    freq_out = []
    time_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
        time_out.append(ds.time_out)
    freq_out = np.unique(np.array(freq_out))
    time_out = np.unique(np.array(time_out))

    nband = freq_out.size

    # stitch dirty/psf in apparent scale
    # drop_vars to avoid duplicates in memory
    if 'RESIDUAL' in dds[0]:
        residual = np.stack([ds.RESIDUAL.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('RESIDUAL') for ds in dds]
    else:
        residual = np.stack([ds.DIRTY.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('DIRTY') for ds in dds]
    if 'MODEL' in dds[0]:
        model = np.stack([ds.MODEL.values[0] for ds in dds], axis=0)
        dds = [ds.drop_vars('MODEL') for ds in dds]
    else:
        model = np.zeros((nband, nx, ny))
    psf = np.stack([ds.PSF.values[0] for ds in dds], axis=0)
    psfhat = np.stack([ds.PSFHAT.values[0] for ds in dds], axis=0)
    abspsf = np.stack([np.abs(ds.PSFHAT.values[0]) for ds in dds], axis=0)
    beam = np.stack([ds.BEAM.values[0] for ds in dds], axis=0)
    dds = [ds.drop_vars(('PSF','PSFHAT')) for ds in dds]
    wsums = np.stack([ds.WSUM.values[0] for ds in dds], axis=0)
    fsel = wsums > 0  # keep track of empty bands


    wsum = np.sum(wsums)
    # TODO - do we really need to keep all three versions of the PSF in memory here?
    psf /= wsum
    psfhat /= wsum
    abspsf /= wsum
    residual /= wsum
    residual_mfs = np.sum(residual, axis=0)

    # for intermediary results
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq, casambm=False)
    if 'niters' in dds[0].attrs:
        iter0 = dds[0].niters
    else:
        iter0 = 0

    # TODO - check coordinates match
    # Add option to interp onto coordinates?
    if opts.mask is not None:
        mask = load_fits(opts.mask, dtype=residual.dtype).squeeze()
        assert mask.shape == (nx, ny)
        log.info('Using provided fits mask')
    else:
        mask = np.ones((nx, ny), dtype=residual.dtype)

    # PCG related options for flux mop
    cgopts = {}
    cgopts['tol'] = opts.cg_tol
    cgopts['maxit'] = opts.cg_maxit
    cgopts['minit'] = opts.cg_minit
    cgopts['verbosity'] = opts.cg_verbose
    cgopts['report_freq'] = opts.cg_report_freq
    cgopts['backtrack'] = opts.backtrack

    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    best_rms = rms
    best_rmax = rmax
    best_model = model.copy()
    diverge_count = 0
    if opts.threshold is None:
        threshold = opts.rmsfactor * rms
    else:
        threshold = opts.threshold

    precond = hess_psf(nx, ny, abspsf,
                       beam=mask[None, :, :] * beam,
                       eta=opts.eta*wsums,
                       nthreads=opts.nthreads,
                       cgtol=opts.cg_tol,
                       cgmaxit=opts.cg_maxit,
                       cgverbose=opts.cg_verbose,
                       cgrf=opts.cg_report_freq,
                       taper_width=np.minimum(int(0.1*nx), 32))


    log.info(f"Iter {iter0}: peak residual = {rmax:.3e}, rms = {rms:.3e}")
    for k in range(iter0, iter0 + opts.niter):
        log.info("Cleaning")
        x, status = clark(residual, psf, psfhat, wsums/wsum, mask,
                          threshold=threshold,
                          gamma=opts.gamma,
                          pf=opts.peak_factor,
                          maxit=opts.minor_maxit,
                          subpf=opts.sub_peak_factor,
                          submaxit=opts.subminor_maxit,
                          verbosity=opts.verbose,
                          report_freq=opts.report_freq,
                          nthreads=opts.nthreads)
        model += x

        # write component model
        log.info(f"Writing model at iter {k+1} to "
              f"{basename}_{opts.suffix}_model.mds")
        try:
            coeffs, Ix, Iy, expr, params, texpr, fexpr = \
                fit_image_cube(time_out, freq_out[fsel], model[None, fsel, :, :],
                               wgt=wsums[None, fsel],
                               nbasisf=int(np.sum(fsel)),
                               method='Legendre')
            # save interpolated dataset
            data_vars = {
                'coefficients': (('par', 'comps'), coeffs),
            }
            coords = {
                'location_x': (('x',), Ix),
                'location_y': (('y',), Iy),
                'params': (('par',), params),  # already converted to list
                'times': (('t',), time_out),  # to allow rendering to original grid
                'freqs': (('f',), freq_out)
            }
            attrs = {
                'spec': 'genesis',
                'cell_rad_x': cell_rad,
                'cell_rad_y': cell_rad,
                'npix_x': nx,
                'npix_y': ny,
                'texpr': texpr,
                'fexpr': fexpr,
                'center_x': dds[0].x0,
                'center_y': dds[0].y0,
                'flip_u': dds[0].flip_u,
                'flip_v': dds[0].flip_v,
                'flip_w': dds[0].flip_w,
                'ra': dds[0].ra,
                'dec': dds[0].dec,
                'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
                'parametrisation': expr  # already converted to str
            }

            coeff_dataset = xr.Dataset(data_vars=data_vars,
                                       coords=coords,
                                       attrs=attrs)
            coeff_dataset.to_zarr(f"{basename}_{opts.suffix}_model.mds",
                                  mode='w')
        except Exception as e:
            log.info(f"Exception {e} raised during model fit .")

        save_fits(np.mean(model[fsel], axis=0),
                  fits_oname + f'_{opts.suffix}_model_{k+1}.fits',
                  hdr_mfs)

        log.info(f'Computing residual')
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            resid, _ = compute_residual(ds_name,
                                     nx, ny,
                                     cell_rad, cell_rad,
                                     ds_name,
                                     model[b][None, :, :],  # add back corr axis
                                     nthreads=opts.nthreads,
                                     epsilon=opts.epsilon,
                                     do_wgridding=opts.do_wgridding,
                                     double_accum=opts.double_accum)
            residual[b] = resid[0]  # remove corr axis
        residual /= wsum
        residual_mfs = np.sum(residual, axis=0)
        save_fits(residual_mfs,
                  fits_oname + f'_{opts.suffix}_residual_{k+1}.fits',
                  hdr_mfs)

        # report rms and rmax inside mask
        rmsp = rms
        # tmp_mask = ~np.any(model, axis=0)
        rms = np.std(residual_mfs[mask>0])
        rmax = np.abs(residual_mfs[mask>0]).max()

        # base this on rmax?
        if rms < best_rms:
            best_rms = rms
            best_rmax = rmax
            best_model = model.copy()

        # these are not updated in compute_residual
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            ds = ds.assign(**{
                'MODEL': (('corr', 'x', 'y'), model[b][None, :, :]),
                'MODEL_BEST': (('corr', 'x', 'y'), best_model[b][None, :, :]),
            })
            attrs = {}
            attrs['rms'] = best_rms
            attrs['rmax'] = best_rmax
            attrs['niters'] = k+1
            ds = ds.assign_attrs(**attrs)
            ds.to_zarr(ds_name, mode='a')

        if opts.threshold is None:
            threshold = opts.rmsfactor * rms
        else:
            threshold = opts.threshold

        # trigger flux mop if clean has stalled, not converged or
        # we have reached the final iteration/threshold
        status |= k == iter0 + opts.niter-1
        status |= rmax <= threshold
        if opts.mop_flux and status:
            log.info(f"Extracting flux at iter {k+1}")
            mopmask = np.any(model, axis=0)
            if opts.dirosion:
                struct = ndimage.generate_binary_structure(2, opts.dirosion)
                mopmask = ndimage.binary_dilation(mopmask, structure=struct)
                mopmask = ndimage.binary_erosion(mopmask, structure=struct)
            x0 = np.zeros_like(x)
            # x0[:, mopmask] = residual_mfs[mopmask]
            # TODO - applying mask as beam is wasteful
            mopmask = (mopmask.astype(residual.dtype) * mask)[None, :, :]
            precond.set_beam(mopmask * beam)
            x = precond.idot(residual * beam,
                             mode='psf',
                             init_x0=False)
            model += opts.mop_gamma*x

            log.info(f'Computing residual')
            for ds_name, ds in zip(dds_list, dds):
                b = int(ds.bandid)
                resid, _ = compute_residual(ds_name,
                                        nx, ny,
                                        cell_rad, cell_rad,
                                        ds_name,
                                        model[b][None, :, :],  # add back corr axis
                                        nthreads=opts.nthreads,
                                        epsilon=opts.epsilon,
                                        do_wgridding=opts.do_wgridding,
                                        double_accum=opts.double_accum)
                residual[b] = resid[0]  # remove corr axis
            residual /= wsum
            residual_mfs = np.sum(residual, axis=0)

            save_fits(residual_mfs,
                      f'{fits_oname}_{opts.suffix}_postmop{k+1}_residual_mfs.fits',
                      hdr_mfs)

            save_fits(np.mean(model[fsel], axis=0),
                      f'{fits_oname}_{opts.suffix}_postmop{k+1}_model_mfs.fits',
                      hdr_mfs)

            rmsp = rms
            # tmp_mask = ~np.any(model, axis=0)
            rms = np.std(residual_mfs[mask>0])
            rmax = np.abs(residual_mfs[mask>0]).max()

            # base this on rmax?
            if rms < best_rms:
                best_rms = rms
                best_rmax = rmax
                best_model = model.copy()
                for ds_name, ds in zip(dds_list, dds):
                    b = int(ds.bandid)
                    ds = ds.assign(**{
                        'MODEL_BEST': (('corr', 'x', 'y'), best_model[b][None, :, :])
                    })

                    attrs = {}
                    attrs['rms'] = best_rms
                    attrs['rmax'] = best_rmax
                    attrs['niters'] = k+1
                    ds = ds.assign_attrs(**attrs)
                    ds.to_zarr(ds_name, mode='a')
                    ds.to_zarr(ds_name, mode='a')

            if opts.threshold is None:
                threshold = opts.rmsfactor * rms
            else:
                threshold = opts.threshold

        log.info(f"Iter {k+1}: peak residual = {rmax:.3e}, rms = {rms:.3e}")

        if rmax <= threshold:
            log.info("Terminating because final threshold has been reached")
            break

        if rms > rmsp:
            diverge_count += 1
            if diverge_count > 3:
                log.info("Algorithm is diverging. Terminating.")
                break

        # keep track of total number of iterations
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            ds = ds.assign_attrs(**{
                'niter': k,
            })

            ds.to_zarr(ds_name, mode='a')

    return


def _fskclean(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    from functools import partial
    import numpy as np
    import xarray as xr
    from pfb.utils.fits import set_wcs, save_fits, load_fits
    from pfb.deconv.clark import fsclark
    from pfb.utils.naming import xds_from_url
    from pfb.operators.gridder import compute_residual
    from pfb.operators.hessian import fshessian_jax
    from scipy import ndimage
    from pfb.utils.modelspec import fit_image_fscube
    from jax.scipy.sparse.linalg import cg

    basename = opts.output_filename
    if opts.fits_output_folder is not None:
        fits_oname = opts.fits_output_folder + '/' + basename.split('/')[-1]
    else:
        fits_oname = basename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds, dds_list = xds_from_url(dds_name)

    nx, ny = dds[0].x.size, dds[0].y.size
    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf
    freq_out = []
    time_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
        time_out.append(ds.time_out)
    freq_out = np.unique(np.array(freq_out))
    time_out = np.unique(np.array(time_out))

    nband = freq_out.size
    ncorr = dds[0].corr.size

    # stitch dirty/psf in apparent scale
    # drop_vars to avoid duplicates in memory
    if 'RESIDUAL' in dds[0]:
        residual = np.stack([ds.RESIDUAL.values for ds in dds], axis=0)
        dds = [ds.drop_vars('RESIDUAL') for ds in dds]
    else:
        residual = np.stack([ds.DIRTY.values for ds in dds], axis=0)
        dds = [ds.drop_vars('DIRTY') for ds in dds]
    if 'MODEL' in dds[0]:
        model = np.stack([ds.MODEL.values for ds in dds], axis=0)
        dds = [ds.drop_vars('MODEL') for ds in dds]
    else:
        model = np.zeros((nband, ncorr, nx, ny))
    psf = np.stack([ds.PSF.values for ds in dds], axis=0)
    psfhat = np.stack([ds.PSFHAT.values for ds in dds], axis=0)
    dds = [ds.drop_vars(['PSF', 'PSFHAT']) for ds in dds]
    wsums = np.stack([ds.WSUM.values for ds in dds], axis=0)

    wsum = np.sum(wsums, axis=0)
    # ensure wsums sum to one for each correlation
    wsums /= wsum[None, :]
    psf /= wsum[None, :, None, None]
    psfhat /= wsum[None, :, None, None]
    residual /= wsum[None, :, None, None]
    residual_mfs = np.sum(residual, axis=0)

    # for intermediary results
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq, casambm=False)
    if 'niters' in dds[0].attrs:
        iter0 = dds[0].niters
    else:
        iter0 = 0

    # TODO - check coordinates match
    # Add option to interp onto coordinates?
    if opts.mask is not None:
        mask = load_fits(opts.mask, dtype=residual.dtype).squeeze()
        assert mask.shape == (nx, ny)
        log.info('Using provided fits mask')
    else:
        mask = np.ones((nx, ny), dtype=residual.dtype)

    # PCG related options for flux mop
    cgopts = {}
    cgopts['tol'] = opts.cg_tol
    cgopts['maxit'] = opts.cg_maxit
    cgopts['minit'] = opts.cg_minit
    cgopts['verbosity'] = opts.cg_verbose
    cgopts['report_freq'] = opts.cg_report_freq


    rms = np.std(residual_mfs, axis=(-2,-1)).max()
    rmax = np.abs(residual_mfs).max()
    best_rms = rms
    best_rmax = rmax
    best_model = model.copy()
    diverge_count = 0
    if opts.threshold is None:
        threshold = opts.rmsfactor * rms
    else:
        threshold = opts.threshold

    log.info(f"Iter {iter0}: peak residual = {rmax:.3e}, rms = {rms:.3e}")
    for k in range(iter0, iter0 + opts.niter):
        log.info("Cleaning")
        x, status = fsclark(residual, psf, psfhat,
                            wsums, mask,
                            threshold=threshold,
                            gamma=opts.gamma,
                            pf=opts.peak_factor,
                            maxit=opts.minor_maxit,
                            subpf=opts.sub_peak_factor,
                            submaxit=opts.subminor_maxit,
                            verbosity=opts.verbose,
                            report_freq=opts.report_freq,
                            nthreads=opts.nthreads)
        model += x

        # write component model
        log.info(f"Writing model at iter {k+1} to "
              f"{basename}_{opts.suffix}_model.mds")
        try:
            coeffs, Ix, Iy, expr, params, texpr, fexpr = \
                fit_image_fscube(freq_out, model,
                                 wgt=wsums,
                                 nbasisf=int(nband),
                                 method='Legendre')
            # save interpolated dataset
            data_vars = {
                'coefficients': (('corr', 'par', 'comps'), coeffs),
            }
            coords = {
                'location_x': (('x',), Ix),
                'location_y': (('y',), Iy),
                'params': (('par',), params),  # already converted to list
                'times': (('t',), time_out),  # to allow rendering to original grid
                'freqs': (('f',), freq_out),
                'corr': (('corr',), dds[0].corr.values),
            }
            attrs = {
                'spec': 'genesis',
                'cell_rad_x': cell_rad,
                'cell_rad_y': cell_rad,
                'npix_x': nx,
                'npix_y': ny,
                'texpr': texpr,
                'fexpr': fexpr,
                'center_x': dds[0].x0,
                'center_y': dds[0].y0,
                'flip_u': dds[0].flip_u,
                'flip_v': dds[0].flip_v,
                'flip_w': dds[0].flip_w,
                'ra': dds[0].ra,
                'dec': dds[0].dec,
                'parametrisation': expr  # already converted to str
            }

            coeff_dataset = xr.Dataset(data_vars=data_vars,
                                       coords=coords,
                                       attrs=attrs)
            coeff_dataset.to_zarr(f"{basename}_{opts.suffix}_model.mds",
                                  mode='w')
        except Exception as e:
            log.info(f"Exception {e} raised during model fit .")

        save_fits(np.mean(model, axis=0),
                  fits_oname + f'_{opts.suffix}_model_{k+1}.fits',
                  hdr_mfs)

        log.info(f'Computing residual')
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            resid, _ = compute_residual(ds_name,
                                     nx, ny,
                                     cell_rad, cell_rad,
                                     ds_name,
                                     model[b],
                                     nthreads=opts.nthreads,
                                     epsilon=opts.epsilon,
                                     do_wgridding=opts.do_wgridding,
                                     double_accum=opts.double_accum)
            residual[b] = resid
        residual /= wsum[None, :, None, None]
        residual_mfs = np.sum(residual, axis=0)
        save_fits(residual_mfs,
                  fits_oname + f'_{opts.suffix}_residual_{k+1}.fits',
                  hdr_mfs)

        # report rms and rmax inside mask
        rmsp = rms
        rms = np.std(residual_mfs, axis=(-2,-1)).max()
        rmax = np.abs(residual_mfs).max()

        # base this on rmax?
        if rms < best_rms:
            best_rms = rms
            best_rmax = rmax
            best_model = model.copy()

        # trigger flux mop if clean has stalled, not converged or
        # we have reached the final iteration/threshold
        status |= k == iter0 + opts.niter-1
        status |= rmax <= threshold
        if opts.mop_flux and status:
            log.info(f"Extracting flux at iter {k+1}")
            mopmask = np.any(model, axis=(0,1))
            if opts.dirosion:
                struct = ndimage.generate_binary_structure(2, opts.dirosion)
                mopmask = ndimage.binary_dilation(mopmask, structure=struct)
                mopmask = ndimage.binary_erosion(mopmask, structure=struct)
            # TODO - apply beam/n with mask
            mopmask = (mopmask.astype(residual.dtype) * mask)
            A = partial(fshessian_jax, nx, ny, nx_psf, ny_psf, rmax,
                        mopmask, np.abs(psfhat))
            x, _ = cg(A,
                      residual*mopmask[None, None, :, :],
                      tol=opts.cg_tol,
                      maxiter=opts.cg_maxit)

            model += opts.mop_gamma*np.array(x)

            log.info(f'Computing residual')
            for ds_name, ds in zip(dds_list, dds):
                b = int(ds.bandid)
                resid, _ = compute_residual(ds_name,
                                        nx, ny,
                                        cell_rad, cell_rad,
                                        ds_name,
                                        model[b],
                                        nthreads=opts.nthreads,
                                        epsilon=opts.epsilon,
                                        do_wgridding=opts.do_wgridding,
                                        double_accum=opts.double_accum)
                residual[b] = resid
            residual /= wsum[None, :, None, None]
            residual_mfs = np.sum(residual, axis=0)

            save_fits(residual_mfs,
                      f'{fits_oname}_{opts.suffix}_postmop{k+1}_residual_mfs.fits',
                      hdr_mfs)

            save_fits(np.mean(model, axis=0),
                      f'{fits_oname}_{opts.suffix}_postmop{k+1}_model_mfs.fits',
                      hdr_mfs)

            rmsp = rms
            rms = np.std(residual_mfs, axis=(-2,-1)).max()
            rmax = np.abs(residual_mfs).max()

            # base this on rmax?
            if rms < best_rms:
                best_rms = rms
                best_rmax = rmax
                best_model = model.copy()

        # these are not updated in compute_residual
        for ds_name, ds in zip(dds_list, dds):
            b = int(ds.bandid)
            if (model==best_model).all():
                ds = ds.assign(**{
                    'MODEL_BEST': (('corr', 'x', 'y'), best_model[b]),
                })
            attrs = {}
            attrs['rms'] = best_rms
            attrs['rmax'] = best_rmax
            attrs['niters'] = k+1
            ds = ds.assign_attrs(**attrs)
            # we don't want to write everything to disk
            dsw = ds[['MODEL_BEST']]
            dsw.to_zarr(ds_name, mode='a')

        if opts.threshold is None:
            threshold = opts.rmsfactor * rms
        else:
            threshold = opts.threshold

        log.info(f"Iter {k+1}: peak residual = {rmax:.3e}, rms = {rms:.3e}")

        if rmax <= threshold:
            log.info("Terminating because final threshold has been reached")
            break

        if rms > rmsp:
            diverge_count += 1
            if diverge_count > 3:
                log.info("Algorithm is diverging. Terminating.")
                break

    return



import numpy as np
import dask.array as da
import sys
from africanus.model.spi.dask import fit_spi_components


def fit_spi(image: np.ndarray,
            beam: np.ndarray,
            freqs: np.ndarray,
            weights: np.ndarray,
            threshold: float,
            nthreads: int,
            pb_min: float,
            padding_frac: float,
            ref_freq: float,
            dest=sys.stdout):

    try:
        assert image.ndim == 3
        assert beam.ndim == 3
        assert image.shape == beam.shape
        assert image.shape[0] > 1
        assert freqs.size == image.shape[0]
        assert weights.size == image.shape[0]
    except Exception as e:
        raise e

    # beam cut off
    image = np.where(beam > pb_min, image, 0)

    nband = image.shape[0]

    # get pixels above threshold
    minimage = np.amin(image, axis=0)
    maskindices = np.argwhere(minimage > threshold)
    if not maskindices.size:
        raise ValueError("No components found above threshold. "
                         "Try lowering your threshold."
                         "Max of convolved model is %3.2e" % image.max())
    fitcube = image[:, maskindices[:, 0], maskindices[:, 1]].T
    beam_comps = beam[:, maskindices[:, 0], maskindices[:, 1]].T

    ncomps, _ = fitcube.shape
    fitcube = da.from_array(fitcube.astype(np.float64),
                            chunks=(ncomps//nthreads, nband))
    beam_comps = da.from_array(beam_comps.astype(np.float64),
                               chunks=(ncomps//nthreads, nband))
    weights = da.from_array(weights.astype(np.float64), chunks=(nband))
    freqsdask = da.from_array(freqs.astype(np.float64), chunks=(nband))

    print("Fitting %i components" % ncomps, file=dest)
    alpha, alpha_err, Iref, i0_err = \
        fit_spi_components(fitcube, weights, freqsdask, ref_freq,
                           beam=beam_comps).compute()
    print("Done. Writing output. \n", file=dest)

    alphamap = np.zeros(image[0].shape, dtype=image.dtype)
    alphamap[...] = np.nan
    alpha_err_map = np.zeros(image[0].shape, dtype=image.dtype)
    alpha_err_map[...] = np.nan
    i0map = np.zeros(image[0].shape, dtype=image.dtype)
    i0map[...] = np.nan
    i0_err_map = np.zeros(image[0].shape, dtype=image.dtype)
    i0_err_map[...] = np.nan
    alphamap[maskindices[:, 0], maskindices[:, 1]] = alpha
    alpha_err_map[maskindices[:, 0], maskindices[:, 1]] = alpha_err
    i0map[maskindices[:, 0], maskindices[:, 1]] = Iref
    i0_err_map[maskindices[:, 0], maskindices[:, 1]] = i0_err

    return alphamap, alpha_err_map, i0map, i0_err_map



import os
import sys
import logging

def set_envs(nthreads, ncpu):
    os.environ["OMP_NUM_THREADS"] = str(nthreads)
    os.environ["OPENBLAS_NUM_THREADS"] = str(nthreads)
    os.environ["MKL_NUM_THREADS"] = str(nthreads)
    os.environ["VECLIB_MAXIMUM_THREADS"] = str(nthreads)
    os.environ["NPY_NUM_THREADS"] = str(nthreads)
    os.environ["NUMBA_NUM_THREADS"] = str(nthreads)
    # os.environ["JAX_PLATFORMS"] = 'cpu'
    # os.environ["XLA_FLAGS"] = ("--xla_cpu_multi_thread_eigen=true "
    #                            "--xla_force_host_platform_device_count=4 "
    #                            f"intra_op_parallelism_threads={str(nthreads)}")
    os.environ["JAX_ENABLE_X64"] = 'True'
    # this may be required for numba parallelism
    # find python and set LD_LIBRARY_PATH
    paths = sys.path
    ppath = [paths[i] for i in range(len(paths)) if 'pfb/bin' in paths[i]]
    if len(ppath):
        ldpath = ppath[0].replace('bin', 'lib')
        ldcurrent = os.environ.get('LD_LIBRARY_PATH', '')
        os.environ["LD_LIBRARY_PATH"] = f'{ldpath}:{ldcurrent}'
        # TODO - should we fall over in else?

    ne_threads = min(ncpu, nthreads)
    os.environ["NUMEXPR_NUM_THREADS"] = str(ne_threads)


def set_client(nworkers, log, stack=None, host_address=None,
               direct_to_workers=False, client_log_level=None):

    import warnings
    warnings.filterwarnings("ignore", message="Port 8787 is already in use")
    if client_log_level == 'error':
        logging.getLogger("distributed").setLevel(logging.ERROR)
        logging.getLogger("bokeh").setLevel(logging.ERROR)
        logging.getLogger("tornado").setLevel(logging.CRITICAL)
    elif client_log_level == 'warning':
        logging.getLogger("distributed").setLevel(logging.WARNING)
        logging.getLogger("bokeh").setLevel(logging.WARNING)
        logging.getLogger("tornado").setLevel(logging.WARNING)
    elif client_log_level == 'info':
        logging.getLogger("distributed").setLevel(logging.INFO)
        logging.getLogger("bokeh").setLevel(logging.INFO)
        logging.getLogger("tornado").setLevel(logging.INFO)
    elif client_log_level == 'debug':
        logging.getLogger("distributed").setLevel(logging.DEBUG)
        logging.getLogger("bokeh").setLevel(logging.DEBUG)
        logging.getLogger("tornado").setLevel(logging.DEBUG)

    import dask
    # set up client
    host_address = host_address or os.environ.get("DASK_SCHEDULER_ADDRESS")
    if host_address is not None:
        from distributed import Client
        log.info("Initialising distributed client.")
        if stack is not None:
            client = stack.enter_context(Client(host_address))
        else:
            client = Client(host_address)
    else:
        from dask.distributed import Client, LocalCluster
        log.info("Initialising client with LocalCluster.")
        dask.config.set({
                'distributed.comm.compression': {
                    'on': True,
                    'type': 'lz4'
                }
        })
        cluster = LocalCluster(processes=True,
                               n_workers=nworkers,
                               threads_per_worker=1,
                               memory_limit=0,  # str(mem_limit/nworkers)+'GB'
                               asynchronous=False)
        if stack is not None:
            cluster = stack.enter_context(cluster)
        client = Client(cluster,
                        direct_to_workers=direct_to_workers)
        if stack is not None:
            client = stack.enter_context(client)

    client.wait_for_workers(nworkers)
    dashboard_url = client.dashboard_link
    log.info(f"Dask Dashboard URL at {dashboard_url}")

    return client


def logo():
    print("""
    ███████████  ███████████ ███████████
   ░░███░░░░░███░░███░░░░░░█░░███░░░░░███
    ░███    ░███ ░███   █ ░  ░███    ░███
    ░██████████  ░███████    ░██████████
    ░███░░░░░░   ░███░░░█    ░███░░░░░███
    ░███         ░███  ░     ░███    ░███
    █████        █████       ███████████
   ░░░░░        ░░░░░       ░░░░░░░░░░░
    """)



import numpy as np

def generate_time_profile(times, peak_time, duration, shape):
    """Generate time profile for a single pulse"""
    t0 = peak_time
    
    if shape == 'gaussian':
        sigma = duration
        profile = np.exp(-(times - t0)**2 / (2*sigma**2))
        
    elif shape == 'exponential':
        tau = duration
        profile = np.where(times >= t0, np.exp(-(times - t0) / tau), 0.0)
        
    elif shape == 'step':
        delta_t = duration
        profile = np.where((times >= t0) & (times <= t0 + delta_t), 1.0, 0.0)
        
    else:
        raise ValueError(f"Unknown time profile shape: {shape}")
    
    return profile

def generate_frequency_profile(freqs, peak_flux, reference_freq, spectral_index):
    """Generate frequency profile (power law spectrum)"""
    return peak_flux * (freqs / reference_freq)**spectral_index

def generate_transient_spectra(times, freqs, transient_params):
    """
    Generate dynamic spectra for a single transient
    
    Parameters:
    -----------
    times : numpy.ndarray
        Time array (seconds)
    freqs : numpy.ndarray  
        Frequency array (Hz)
    transient_params : dict
        Transient parameters from YAML
        
    Returns:
    --------
    spectra : numpy.ndarray
        Dynamic spectra with shape (ntime, nfreq)
    """
    # Extract parameters
    time_params = transient_params['time']
    freq_params = transient_params['frequency']
    periodicity = transient_params.get('periodicity', {'enabled': False})
    
    # since peak_time is set from start of observation
    times = times - times[0]

    # Generate base time profile
    time_profile = generate_time_profile(
        times, 
        time_params['peak_time'], 
        time_params['duration'], 
        time_params['shape']
    )
    
    # Add periodicity if enabled
    if periodicity.get('enabled', False):
        period = periodicity['period']
        total_duration = periodicity.get('total_duration', times[-1] - times[0])
        
        # Create periodic version by summing shifted copies
        periodic_profile = np.zeros_like(times)
        t_start = time_params['peak_time']
        
        # Calculate how many periods fit in the total duration
        n_periods = int((total_duration - t_start) / period) + 1
        
        for n in range(n_periods):
            t_shift = n * period
            if t_start + t_shift <= total_duration:
                shifted_profile = generate_time_profile(
                    times, 
                    time_params['peak_time'] + t_shift, 
                    time_params['duration'], 
                    time_params['shape']
                )
                periodic_profile += shifted_profile
        
        time_profile = periodic_profile
    
    # Generate frequency profile
    freq_profile = generate_frequency_profile(
        freqs,
        freq_params['peak_flux'],
        freq_params['reference_freq'], 
        freq_params['spectral_index']
    )
    
    return time_profile, freq_profile


import concurrent.futures as cf
import numpy as np
import pywt
from pfb.wavelets.wavelets_jsk import (get_buffer_size,
                                       dwt2d, idwt2d)


def dwt(x, buffer, dec_lo, dec_hi, nlevel, i):
    dwt2d(x, buffer, dec_lo, dec_hi, nlevel)
    return buffer, i


def idwt(buffer, rec_lo, rec_hi, nlevel, nx, ny, i):
    image = np.zeros((nx, ny), dtype=buffer.dtype)
    idwt2d(buffer, image, rec_lo, rec_hi, nlevel)
    return image, i


class psi_band(object):
    def __init__(self, bases, nlevel, nx, ny, nthreads):
        self.bases = bases
        self.nbasis = len(bases)
        self.nlevel = nlevel
        self.nx = nx
        self.ny = ny
        self.nthreads = nthreads
        self.dec_lo ={}
        self.dec_hi ={}
        self.rec_lo ={}
        self.rec_hi ={}
        self.buffer_size = {}
        self.buffer = {}
        self.nmax = 0
        for wavelet in bases:
            if wavelet=='self':
                self.buffer_size[wavelet] = nx*ny
                self.nmax = np.maximum(self.nmax, self.buffer_size[wavelet])
                continue
            # Get the required filter banks from pywt.
            wvlt = pywt.Wavelet(wavelet)
            self.dec_lo[wavelet] = np.array(wvlt.filter_bank[0])  # Low pass, decomposition.
            self.dec_hi[wavelet] = np.array(wvlt.filter_bank[1])  # Hi pass, decomposition.
            self.rec_lo[wavelet] = np.array(wvlt.filter_bank[2])  # Low pass, recon.
            self.rec_hi[wavelet] = np.array(wvlt.filter_bank[3])  # Hi pass, recon.

            self.buffer_size[wavelet] = get_buffer_size((nx, ny),
                                                        self.dec_lo[wavelet].size,
                                                        nlevel)
            self.buffer[wavelet] = np.zeros(self.buffer_size[wavelet],
                                            dtype=np.float64)

            self.nmax = np.maximum(self.nmax, self.buffer_size[wavelet])


    def dot(self, x):
        '''
        signal to coeffs

        Input:
            x       - (nx, ny) input signal

        Output:
            alpha   - (nbasis, Nnmax) output coeffs
        '''
        alpha = np.zeros((self.nbasis, self.nmax),
                         dtype=x.dtype)

        # comment below to eliminate the possibility of
        # wavelets/cf.futures being the culprit
        # run with bases=self only
        alpha[0, 0:self.buffer_size[self.bases[0]]] = x.ravel()

        # with cf.ThreadPoolExecutor(max_workers=self.nthreads) as executor:
        #     futures = []
        #     for i, wavelet in enumerate(self.bases):
        #         if wavelet=='self':
        #             alpha[i, 0:self.buffer_size[wavelet]] = x.ravel()
        #             continue

        #         f = executor.submit(dwt, x,
        #                             self.buffer[wavelet],
        #                             self.dec_lo[wavelet],
        #                             self.dec_hi[wavelet],
        #                             self.nlevel,
        #                             i)

        #         futures.append(f)

        #     for f in cf.as_completed(futures):
        #         buffer, i = f.result()
        #         alpha[i, 0:self.buffer_size[self.bases[i]]] = buffer

        return alpha

    def hdot(self, alpha):
        '''
        coeffs to signal

        Input:
            alpha   - (nbasis, Nxmax, Nymax) per basis output coeffs

        Output:
            x       - (nx, ny) output signal
        '''
        nx = self.nx
        ny = self.ny
        x = np.zeros((nx, ny), dtype=alpha.dtype)

        # comment below to eliminate the possibility of
        # wavelets/cf.futures being the culprit
        # run with bases=self only
        nmax = self.buffer_size[self.bases[0]]
        x = alpha[0, 0:nmax].reshape(nx, ny)

        # with cf.ThreadPoolExecutor(max_workers=self.nthreads) as executor:
        #     futures = []
        #     for i, wavelet in enumerate(self.bases):
        #         nmax = self.buffer_size[wavelet]
        #         if wavelet=='self':
        #             x += alpha[i, 0:nmax].reshape(nx, ny)
        #             continue

        #         f = executor.submit(idwt,
        #                             alpha[i, 0:nmax],
        #                             self.rec_lo[wavelet],
        #                             self.rec_hi[wavelet],
        #                             self.nlevel,
        #                             self.nx,
        #                             self.ny,
        #                             i)


        #         futures.append(f)

        #     for f in cf.as_completed(futures):
        #         image, i = f.result()
        #         x += image

        return x



import numpy as np
from distributed import get_client
from scipy.linalg import norm
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('PM')


def power_method(
        A,
        imsize,
        b0=None,
        tol=1e-5,
        maxit=250,
        verbosity=1,
        report_freq=25):
    if b0 is None:
        b = np.random.randn(*imsize)
        b /= norm(b)
    else:
        b = b0 / norm(b0)
    beta = 1.0
    eps = 1.0
    k = 0
    bp = b.copy()
    while eps > tol and k < maxit:
        b = A(bp)
        bnorm = np.linalg.norm(b)
        betap = beta
        beta = np.vdot(bp, b) / np.vdot(bp, bp)
        b /= bnorm
        # this is a scalar
        eps = np.linalg.norm(beta - betap) / betap
        k += 1
        bp[...] = b[...]

        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} eps = {eps:.3e}")

    if k == maxit:
        if verbosity:
            log.info(f"Maximum iterations reached. "
                  f"eps = {eps:.3e}, beta = {beta:.3e}")
    else:
        if verbosity:
            log.info(f"Success, converged after {k} iterations. "
                  f"beta = {beta:.3e}")
    return beta, b


def power(A, bp, bnorm, eta):
    bp /= bnorm
    b = A(bp, eta)
    bsumsq = np.sum(b**2)
    beta_num = np.vdot(b, bp)
    beta_den = np.vdot(bp, bp)

    return b, bsumsq, beta_num, beta_den

def sumsq(b):
    return np.sum(b**2)

def bnormf(bsumsq):
    return np.sqrt(np.sum(bsumsq))

def betaf(beta_num, beta_den):
    return np.sum(beta_num)/np.sum(beta_den)

def power_method_dist(actors,
                      nx,
                      ny,
                      nband,
                      tol=1e-4,
                      maxit=200,
                      report_freq=10,
                      verbosity=1,):

    client = get_client()

    bssq = list(map(lambda a: a.init_random(), actors))
    # custom gather?
    bssq = list(map(lambda o: o.result(), bssq))
    bnorm = np.sqrt(np.sum(bssq))
    beta = 1
    for k in range(maxit):
        futures = list(map(lambda a: a.pm_update(bnorm), actors))

        results = list(map(lambda f: f.result(), futures))
        # what is wrong here?
        # bssq = list(map(getitem, results, 0))
        bssq = [r[0] for r in results]
        bnum = [r[1] for r in results]
        bden = [r[2] for r in results]

        bnorm = np.sqrt(np.sum(bssq))
        betap = beta
        beta = np.sum(bnum)/np.sum(bden)

        eps = np.abs(betap - beta)/betap
        if eps < tol:
            break

        if not k % report_freq and verbosity > 1:
            log.info(f"At iteration {k} eps = {eps:.3e}")

    return beta



import numpy as np
from numba import njit, prange
from ducc0.fft import r2c, c2r
from africanus.gps.kernels import exponential_squared as expsq
from pfb.utils.misc import kron_matvec

iFs = np.fft.ifftshift
Fs = np.fft.fftshift


@njit(parallel=True, nogil=True, cache=True, inline='always')
def freqmul(A, x):
    nchan, npix = x.shape
    out = np.zeros((nchan, npix), dtype=x.dtype)
    for i in prange(npix):
        for j in range(nchan):
            for k in range(nchan):
                out[j, i] += A[j, k] * x[k, i]
    return out


@njit(parallel=True, nogil=True, cache=True, inline='always')
def make_kernel(nx_psf, ny_psf, sigma0, length_scale):
    K = np.zeros((1, nx_psf, ny_psf), dtype=np.float64)
    for j in range(nx_psf):
        for k in range(ny_psf):
            l = float(j - (nx_psf // 2))
            m = float(k - (ny_psf // 2))
            K[0, j, k] = sigma0**2 * np.exp(-(l**2 + m**2) / (2 * length_scale**2))
    return K


class mock_array(object):
    def __init__(self, n):
        self.n = n

    @property
    def size(self):
        return self.n**2

    @property
    def shape(self):
        return (self.n, self.n)

    @staticmethod
    def dot(x):
        return x


class Gauss(object):
    def __init__(self, sigma0, nband, nx, ny, nthreads=8):
        self.nthreads = nthreads
        self.nx = nx
        self.ny = ny
        nx_psf = 2*self.nx
        npad_x = (nx_psf - nx)//2
        ny_psf = 2*self.ny
        npad_y = (ny_psf - ny)//2
        self.padding = ((0, 0), (npad_x, npad_x), (npad_y, npad_y))
        self.ax = (1, 2)

        self.unpad_x = slice(npad_x, -npad_x)
        self.unpad_y = slice(npad_y, -npad_y)
        self.lastsize = ny + np.sum(self.padding[-1])

        # set length scales
        length_scale = 0.5

        K = make_kernel(nx_psf, ny_psf, sigma0, length_scale)

        self.K = K
        K_pad = iFs(self.K, axes=self.ax)
        self.Khat = r2c(K_pad, axes=self.ax, forward=True,
                        nthreads=nthreads, inorm=0)
        self.Khatinv = np.where(self.Khat.real > 1e-14, 1.0/self.Khat, 1e-14)

        # get covariance in each dimension
        # pixel coordinates
        self.Kv = mock_array(nband)  # np.eye(nband) * sigma0**2
        self.Kvinv = mock_array(nband)  # np.eye(nband) / sigma0**2
        if nx == ny:
            l_coord = m_coord = np.arange(-(nx//2), nx//2)
            self.Kl = self.Km = expsq(l_coord, l_coord, 1.0, length_scale)
            self.Klinv = self.Kminv = np.linalg.pinv(
                self.Kl, hermitian=True, rcond=1e-12)
            self.Kl *= sigma0**2
            self.Klinv /= sigma0**2
        else:
            l_coord = np.arange(-(nx//2), nx//2)
            m_coord = np.arange(-(ny//2), ny//2)

            self.Kl = expsq(l_coord, l_coord, sigma0, length_scale)
            self.Km = expsq(m_coord, m_coord, 1.0, length_scale)
            self.Klinv = np.linalg.pinv(self.Kl, hermitian=True, rcond=1e-12)
            self.Kminv = np.linalg.pinv(self.Km, hermitian=True, rcond=1e-12)

        # Kronecker matrices for "fast" matrix vector products
        self.Kkron = (self.Kv, self.Kl, self.Km)
        self.Kinvkron = (self.Kvinv, self.Klinv, self.Kminv)

    def convolve(self, x):
        xhat = iFs(np.pad(x, self.padding, mode='constant'), axes=self.ax)
        xhat = r2c(xhat, axes=self.ax, nthreads=self.nthreads,
                   forward=True, inorm=0)
        xhat = c2r(xhat * self.Khat, axes=self.ax, forward=False,
                   lastsize=self.lastsize, inorm=2, nthreads=self.nthreads)
        res = Fs(xhat, axes=self.ax)[:, self.unpad_x, self.unpad_y]
        return res

    def idot(self, x):
        return kron_matvec(self.Kinvkron, x.flatten()).reshape(*x.shape)

    def dot(self, x):
        return kron_matvec(self.Kkron, x.flatten()).reshape(*x.shape)



import numpy as np
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('FISTA')


def back_track_func(x, xp, gradp, likp, L):
    df = x - xp
    return likp + np.vdot(gradp, df) + L * np.vdot(df, df) / 2


def fista(x0,
          L,       # spectral norm of measurement operator
          fprime,  # function returning value and gradient
          prox,    # function implementing prox of regulariser
          tol=1e-3,
          maxit=100,
          report_freq=50,
          verbosity=1):

    # start iterations
    t = 1.0
    x = x0.copy()
    y = x0.copy()
    eps = 1.0
    k = 0
    fidp, gradp = fprime(x)
    for k in range(maxit):
        xp = x.copy()

        # gradient update
        x = y - gradp / L

        # prox w.r.t r(x)
        x = prox(x)

        fidn, gradn = fprime(x)
        # flam = back_track_func(x, xp, gradp, fidp, L)
        i = 0
        while fidn > fidp and i < 10:
            L *= 2.0
            if verbosity > 1:
                log.info("Step size too large, adjusting %f" % L)

            # gradient update
            x = y - gradp / L

            # prox
            x = prox(x)

            fidn, gradn = fprime(x)
            # flam = back_track_func(x, xp, gradp, fidp, L)
            i += 1

        if i == 10:
            if verbosity > 1:
                log.info("Stalled")
            k = maxit - 1
            break

        # fista update
        tp = t
        t = (1. + np.sqrt(1. + 4 * tp**2)) / 2.
        gamma = (tp - 1) / t
        y = x + gamma * (x - xp)

        # convergence check
        eps = np.linalg.norm(x - xp) / np.linalg.norm(x)
        if eps < tol:
            break

        fidp = fidn
        gradp = gradn

        if not k % report_freq and verbosity > 1:
            log.info("At iteration %i eps = %f" % (k, eps))

    if k == maxit - 1:
        if verbosity:
            log.info("Maximum iterations reached. "
                  "Relative difference between updates = %f" %
                  eps)
    else:
        if verbosity:
            log.info("Success, converged after %i iterations" % k)

    return x


fista.__doc__ = r"""
    Algorithm to solve problems of the form

    argmin_x (xbar - x).T A (xbar - x)/2 + lam |x|_21 s.t. x >= 0

    A        - Positive definite Hermitian operator
    xbar     - Data-like term
    x0       - Initial guess
    lam      - Strength of l21 regulariser
    L        - Lipschitz constant of A
    sigma    - l21 step size (set to L/2 by default)
    """



import numpy as np
from scipy.stats import expon
from pfb.opt.power_method import power_method
from pfb.opt.pcg import pcg
from pfb.opt.primal_dual import primal_dual
from pfb.operators.psi import DaskPSI
from pfb.operators.psf import PSF
from pfb.prox.prox_21 import prox_21
from pfb.utils.fits import save_fits
from pfb.utils.misc import Gaussian2D
from pfb.utils import logging as pfb_logging
log = pfb_logging.get_logger('SARA')


def resid_func(x, dirty, hessian, mask, beam, wsum):
    """
    Returns the unattenuated residual
    """
    residual = dirty - hessian(mask(beam(x)))/wsum
    residual_mfs = np.sum(residual, axis=0)
    residual = residual
    return residual, residual_mfs


def sara(psf, model, residual, mask=None, beam_image=None, hessian=None,
         wsum=1, adapt_sig21=True, hdr=None, hdr_mfs=None, outfile=None, cpsf=None,
         nthreads=1, sig_21=1e-6, sigma_frac=100, maxit=10, tol=1e-3,
         gamma=0.99,  psi_levels=2, psi_basis=None, alpha=None,
         pdtol=1e-6, pdmaxit=250, pdverbose=1, positivity=True,
         cgtol=1e-6, cgminit=25, cgmaxit=150, cgverbose=1,
         pmtol=1e-5, pmmaxit=50, pmverbose=1):

    if len(residual.shape) > 3:
        log.error_and_raise("Residual must have shape (nband, nx, ny)",
                            ValueError)

    nband, nx, ny = residual.shape

    if beam_image is None:
        def beam(x): return x
        def beaminv(x): return x
    else:
        try:
            assert beam.shape == (nband, nx, ny)
            def beam(x): return beam_image * x
            def beaminv(x): return np.where(beam_image > 0.01,  x / beam_image, x)
        except BaseException:
            log.error_and_raise("Beam has incorrect shape",
                                ValueError)

    if mask is None:
        def mask(x): return x
    else:
        try:
            if mask.ndim == 2:
                assert mask.shape == (nx, ny)
                def mask(x): return mask[None] * x
            elif mask.ndim == 3:
                assert mask.shape == (1, nx, ny)
                def mask(x): return mask * x
            else:
                raise
        except BaseException:
            log.error_and_raise("Mask has incorrect shape",
                                ValueError)

    # PSF operator
    psfo = PSF(psf, residual.shape, nthreads=nthreads)  #, backward_undersize=1.2)

    if cpsf is None:
        log.error_and_raise("Need to pass in cpsf", ValueError)
    else:
        cpsfo = PSF(cpsf, residual.shape, nthreads=nthreads)

    residual_mfs = np.sum(residual, axis=0)
    residual = mask(beam(residual))
    rmax = np.abs(residual_mfs).max()
    rms = np.std(residual_mfs)

    # wavelet dictionary
    if psi_basis is None:
        psi = DaskPSI(imsize=residual.shape,
                      nlevels=psi_levels, nthreads=nthreads)
    else:
        if not isinstance(psi_basis, list):
            psi_basis = [psi_basis]
        psi = DaskPSI(imsize=residual.shape, nlevels=psi_levels,
                      nthreads=nthreads, bases=psi_basis)

    # set alpha's and sig21's
    # this assumes that the model has been initialised using NNLS
    alpha = np.zeros(psi.nbasis)
    sigmas = np.zeros(psi.nbasis)
    resid_comps = psi.hdot(residual/np.amax(residual.reshape(-1, nx*ny), axis=1)[:, None, None])
    l2_norm = np.linalg.norm(psi.hdot(cpsfo.convolve(model)), axis=1)
    for m in range(psi.nbasis):
        alpha[m] = np.std(resid_comps[m])
        _, sigmas[m] = expon.fit(l2_norm[m], floc=0.0)
        log.info("Basis %i, alpha %f, sigma %f"%(m, alpha[m], sigmas[m]))

    # l21 weights and dual
    weights21 = np.ones((psi.nbasis, psi.nmax), dtype=residual.dtype)
    for m in range(psi.nbasis):
        weights21[m] *= sigmas[m]/sig_21
    dual = np.zeros((psi.nbasis, nband, psi.nmax), dtype=residual.dtype)

    # use PSF to approximate Hessian if not passed in
    if hessian is None:
        hessian = psfo.convolve
        wsum = 1.0

    #  preconditioning operator
    if model.any():
        varmap = np.maximum(rms, sigma_frac * cpsfo.convolve(model))
    else:
        varmap = np.ones(model.shape) * sigma_frac * rms

    def hessf(x):
        # return mask(beam(hessian(mask(beam(x)))))/wsum + x / varmap
        return mask(beam(psfo.convolve(mask(beam(x))))) + x / varmap

    def hessb(x):
        return mask(beam(psfo.convolve(mask(beam(x))))) + x / varmap

    beta, betavec = power_method(hessb, residual.shape, tol=pmtol,
                                 maxit=pmmaxit, verbosity=pmverbose)

    if model.any():
        dirty = residual + hessian(mask(beam(model)))/wsum
    else:
        dirty = residual

    # deconvolve
    for i in range(0, maxit):
        x = pcg(hessf,
                mask(beam(residual)),
                np.zeros_like(residual),
                M=lambda x: x * varmap,
                tol=cgtol, maxit=cgmaxit, minit=cgminit, verbosity=cgverbose)

        # update model
        modelp = model
        model = modelp + gamma * x

        model, dual = primal_dual(hessb, model, modelp, dual, sig_21, psi,
                                  weights21, beta, prox_21, tol=pdtol,
                                  maxit=pdmaxit, report_freq=50, mask=mask,
                                  verbosity=pdverbose, positivity=positivity)

        # get residual
        residual, residual_mfs = resid_func(model, dirty, hessian, mask, beam, wsum)
        model_mfs = np.mean(model, axis=0)
        x_mfs = np.mean(x, axis=0)


        # check stopping criteria
        rmax = np.abs(residual_mfs).max()
        rms = np.std(residual_mfs)
        eps = np.linalg.norm(model - modelp)/np.linalg.norm(model)

        # update variance map (positivity constraint optional)
        varmap = np.maximum(rms, sigma_frac * cpsfo.convolve(model))

        # update spectral norm
        beta, betavec = power_method(hessb, residual.shape, b0=betavec,
                                     tol=pmtol, maxit=pmmaxit,
                                     verbosity=pmverbose)

        log.info("Iter %i: peak residual = %f, rms = %f, eps = %f" % (
              i+1, rmax, rms, eps))

        # reweight
        l2_norm = np.linalg.norm(psi.hdot(model), axis=1)
        for m in range(psi.nbasis):
            if adapt_sig21:
                _, sigmas[m] = expon.fit(l2_norm[m], floc=0.0)
                log.info('basis %i, sigma %f'%sigmas[m])

            weights21[m] = alpha[m]/(alpha[m] + l2_norm[m]) * sigmas[m]/sig_21

        # save current iteration
        if outfile is not None:
            assert hdr is not None
            assert hdr_mfs is not None

            save_fits(outfile + str(i + 1) + '_model_mfs.fits',
                      model_mfs, hdr_mfs)

            save_fits(outfile + str(i + 1) + '_model.fits',
                      model, hdr)

            save_fits(outfile + str(i + 1) + '_update.fits',
                      x, hdr)

            save_fits(outfile + str(i + 1) + '_update_mfs.fits',
                      x_mfs, hdr)

            save_fits(outfile + str(i + 1) + '_residual_mfs.fits',
                      residual_mfs, hdr_mfs)

            save_fits(outfile + str(i + 1) + '_residual.fits',
                      residual*wsum, hdr)


        if eps < tol:
            log.info("Success, convergence after %i iterations" % (i+1))
            break

    return model



import numpy as np
import dask.array as da
from ducc0.fft import r2c, c2r, c2c, good_size
iFs = np.fft.ifftshift
Fs = np.fft.fftshift

def _fft2d_impl(x, nthreads):
    return r2c(iFs(x, axes=(0, 1)), axes=(0, 1), nthreads=nthreads,
               forward=True, inorm=0)

def _fft2d(x, nthreads):
    return _fft2d_impl(x[0], nthreads)

def fft2d(x, nthreads=1):
    if not isinstance(x, da.Array):
        x = da.from_array(x, chunks=(-1, -1), name=False)
    nyp = x.shape[-1]//2 + 1
    return da.blockwise(_fft2d, ('nx', 'nyo2'),
                        x, ('nx', 'ny'),
                        nthreads, None,
                        align_arrays=False,
                        new_axes={'nyo2':nyp},
                        dtype=da.result_type(x, np.complex64))


def _fft_cube_impl(x, nthreads):
    nb, nx, ny = x.shape
    nyp = x.shape[-1]//2 + 1
    xhat = np.zeros((nb, nx, nyp), dtype=np.result_type(x, np.complex64))
    for b in range(nb):
        xhat[b] = r2c(iFs(x[b], axes=(0, 1)), axes=(0, 1), nthreads=nthreads,
                      forward=True, inorm=0)
    return xhat

def _fft_cube(x, nthreads):
    return _fft_cube_impl(x, nthreads)

def fft_cube(x, nthreads=1):
    if not isinstance(x, da.Array):
        x = da.from_array(x, chunks=(1, -1, -1), name=False)
    nyp = x.shape[-1]//2 + 1
    return da.blockwise(_fft_cube, ('nb', 'nx', 'ny'),
                        x, ('nb', 'nx', 'ny'),
                        nthreads, None,
                        align_arrays=False,
                        adjust_chunks={'ny': nyp},
                        dtype=da.result_type(x, np.complex64))



# flake8: noqa
from contextlib import ExitStack
import click
from omegaconf import OmegaConf
from pfb.utils import logging as pfb_logging
from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

log = pfb_logging.get_logger('FLUXTRACTOR')


@click.command(context_settings={'show_default': True})
@clickify_parameters(schema.fluxtractor)
def fluxtractor(**kw):
    '''
    Forward step aka flux mop.

    '''
    opts = OmegaConf.create(kw)

    from pfb.utils.naming import set_output_names
    opts, basedir, oname = set_output_names(opts)

    import psutil
    nthreads = psutil.cpu_count(logical=True)
    ncpu = psutil.cpu_count(logical=False)
    if opts.nthreads is None:
        if opts.nworkers > 1:
            ntpw = nthreads//opts.nworkers
            opts.nthreads = ntpw//2
            ncpu = ntpw//2
        else:
            opts.nthreads = nthreads//2
            ncpu = ncpu//2

    OmegaConf.set_struct(opts, True)

    from pfb import set_envs
    from ducc0.misc import resize_thread_pool
    resize_thread_pool(opts.nthreads)
    set_envs(opts.nthreads, ncpu)

    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    logname = f'{str(opts.log_directory)}/fluxtractor_{timestamp}.log'
    pfb_logging.log_to_file(logname)
    log.info(f'Logs will be written to {logname}')

    pfb_logging.log_options_dict(log, opts)

    from pfb.utils.naming import xds_from_url

    basename = opts.output_filename
    fits_oname = f'{opts.fits_output_folder}/{oname}'
    dds_name =f'{basename}_{opts.suffix}.dds'

    with ExitStack() as stack:
        from distributed import wait
        from pfb import set_client
        if opts.nworkers > 1:
            client = set_client(opts.nworkers, log, stack, client_log_level=opts.log_level)
            from distributed import as_completed
        else:
            log.info("Faking client")
            from pfb.utils.dist import fake_client
            client = fake_client()
            names = [0]
            as_completed = lambda x: x

        ti = time.time()
        _fluxtractor(**opts)

        _, dds_list = xds_from_url(dds_name)

        # convert to fits files
        if opts.fits_mfs or opts.fits_cubes:
            from pfb.utils.fits import dds2fits
            log.info(f"Writing fits files to {fits_oname}_{opts.suffix}")
            futures = []
            fut = client.submit(
                    dds2fits,
                    dds_list,
                    'RESIDUAL_MOPPED',
                    f'{fits_oname}_{opts.suffix}',
                    norm_wsum=True,
                    nthreads=opts.nthreads,
                    do_mfs=opts.fits_mfs,
                    do_cube=opts.fits_cubes)
            futures.append(fut)
            fut = client.submit(
                    dds2fits,
                    dds_list,
                    'MODEL_MOPPED',
                    f'{fits_oname}_{opts.suffix}',
                    norm_wsum=False,
                    nthreads=opts.nthreads,
                    do_mfs=opts.fits_mfs,
                    do_cube=opts.fits_cubes)
            futures.append(fut)
            fut = client.submit(
                    dds2fits,
                    dds_list,
                    'UPDATE',
                    f'{fits_oname}_{opts.suffix}',
                    norm_wsum=False,
                    nthreads=opts.nthreads,
                    do_mfs=opts.fits_mfs,
                    do_cube=opts.fits_cubes)
            futures.append(fut)
            fut = client.submit(
                    dds2fits,
                    dds_list,
                    'X0',
                    f'{fits_oname}_{opts.suffix}',
                    norm_wsum=False,
                    nthreads=opts.nthreads,
                    do_mfs=opts.fits_mfs,
                    do_cube=opts.fits_cubes)
            futures.append(fut)

            for fut in as_completed(futures):
                column = fut.result()
                log.info(f'Done writing {column}')

        log.info(f"All done after {time.time() - ti}s")

def _fluxtractor(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    from itertools import cycle
    import numpy as np
    import xarray as xr
    from pfb.utils.fits import load_fits, set_wcs
    from daskms.fsspec_store import DaskMSStore
    from pfb.utils.naming import xds_from_url
    from pfb.utils.modelspec import fit_image_cube
    from pfb.opt.pcg import pcg_dds

    basename = opts.output_filename

    dds_name = f'{basename}_{opts.suffix}.dds'
    dds_store = DaskMSStore(dds_name)
    dds, dds_list = xds_from_url(dds_store.url)

    nx, ny = dds[0].x.size, dds[0].y.size
    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf
    freq_out = []
    time_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
        time_out.append(ds.time_out)
    freq_out = np.unique(np.array(freq_out))
    time_out = np.unique(np.array(time_out))

    nband = freq_out.size
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)

    if opts.residual_name in dds[0]:
        residual = np.stack([getattr(ds, opts.residual_name).values for ds in dds],
                            axis=0)
    else:
        log.info("Using dirty image as residual")
        residual = np.stack([ds.DIRTY.values for ds in dds], axis=0)
    if opts.model_name in dds[0]:
        model = np.stack([getattr(ds, opts.model_name).values for ds in dds],
                         axis=0)
    else:
        model = np.zeros((nband, nx, ny))
    wsums = np.stack([ds.WSUM.values[0] for ds in dds], axis=0)
    fsel = wsums > 0  # keep track of empty bands
    wsum = np.sum(wsums)
    residual /= wsum
    residual_mfs = np.sum(residual, axis=0)

    # TODO - check coordinates match
    # Add option to interp onto coordinates?
    if opts.mask is not None:
        if opts.mask=='model':
            mask = np.any(model > opts.min_model, axis=0)
            assert mask.shape == (nx, ny)
            mask = mask.astype(residual.dtype)
            log.info('Using model > 0 to create mask')
        else:
            mask = load_fits(opts.mask, dtype=residual.dtype).squeeze()
            assert mask.shape == (nx, ny)
            if opts.or_mask_with_model:
                log.info("Combining model with input mask")
                mask = np.logical_or(mask>0, model_mfs>0).astype(residual.dtype)


            mask = mask.astype(residual.dtype)
            log.info('Using provided fits mask')
    else:
        mask = np.ones((nx, ny), dtype=residual.dtype)
        log.info('Caution - No mask is being applied')

    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    log.info(f"Initial peak residual = {rmax:.3e}, rms = {rms:.3e}")

    log.info("Solving for update")
    try:
        from distributed import get_client, wait, as_completed
        client = get_client()
        names = list(client.scheduler_info()['workers'].keys())
        foo = client.scatter(mask, broadcast=True)
        wait(foo)
    except:
        from pfb.utils.dist import fake_client
        client = fake_client()
        names = [0]
        as_completed = lambda x: x
    futures = []
    for wname, ds_name in zip(cycle(names), dds_list):
        fut = client.submit(
            pcg_dds,
            ds_name,
            opts.eta,
            use_psf=opts.use_psf,
            residual_name=opts.residual_name,
            model_name=opts.model_name,
            mask=mask,
            do_wgridding=opts.do_wgridding,
            epsilon=opts.epsilon,
            double_accum=opts.double_accum,
            nthreads=opts.nthreads,
            zero_model_outside_mask=opts.zero_model_outside_mask,
            tol=opts.cg_tol,
            maxit=opts.cg_maxit,
            verbosity=opts.cg_verbose,
            report_freq=opts.cg_report_freq,
            workers=wname
        )
        futures.append(fut)

    nds = len(futures)
    n_launched = 1
    for fut in as_completed(futures):
        print(f"\rProcessed: {n_launched}/{nds}", end='', flush=True)
        r, b = fut.result()
        residual[b] = r
        n_launched += 1

    print("\n")  # after progressbar above

    residual_mfs = np.sum(residual/wsum, axis=0)
    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    log.info(f"Final peak residual = {rmax:.3e}, rms = {rms:.3e}")

    log.info(f"Writing model to {basename}_{opts.suffix}_model.mds")

    try:
        coeffs, Ix, Iy, expr, params, texpr, fexpr = \
            fit_image_cube(time_out,
                           freq_out[fsel],
                           model[None, fsel, :, :],
                           wgt=wsums[None, fsel],
                           nbasisf=fsel.size,
                           method='Legendre',
                           sigmasq=1e-10)
        # save interpolated dataset
        data_vars = {
            'coefficients': (('par', 'comps'), coeffs),
        }
        coords = {
            'location_x': (('x',), Ix),
            'location_y': (('y',), Iy),
            # 'shape_x':,
            'params': (('par',), params),  # already converted to list
            'times': (('t',), time_out),  # to allow rendering to original grid
            'freqs': (('f',), freq_out)
        }
        mattrs = {
            'spec': 'genesis',
            'cell_rad_x': cell_rad,
            'cell_rad_y': cell_rad,
            'npix_x': nx,
            'npix_y': ny,
            'texpr': texpr,
            'fexpr': fexpr,
            'center_x': dds[0].x0,
            'center_y': dds[0].y0,
            'flip_u': dds[0].flip_u,
            'flip_v': dds[0].flip_v,
            'flip_w': dds[0].flip_w,
            'ra': dds[0].ra,
            'dec': dds[0].dec,
            'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
            'parametrisation': expr  # already converted to str
        }
        for key, val in opts.items():
            if key == 'pd_tol':
                mattrs[key] = pd_tolf
            else:
                mattrs[key] = val

        coeff_dataset = xr.Dataset(data_vars=data_vars,
                            coords=coords,
                            attrs=mattrs)
        coeff_dataset.to_zarr(f"{basename}_{opts.suffix}_model_mopped.mds",
                              mode='w')
    except Exception as e:
            log.info(f"Exception {e} raised during model fit .")

    return



# flake8: noqa
import click
from importlib.metadata import version
from pfb import logo

@click.group()
@click.version_option(version=version("pfb-imaging"), prog_name='pfb')
def cli():
    logo()
    pass

from pfb.workers.init import init
from pfb.workers.grid import grid
from pfb.workers.degrid import degrid
from pfb.workers.kclean import kclean
from pfb.workers.restore import restore
from pfb.workers.model2comps import model2comps
from pfb.workers.fluxtractor import fluxtractor
from pfb.workers.hci import hci
from pfb.workers.smoovie import smoovie
from pfb.workers.sara import sara

# Add all the subcommands to the main CLI group.
for cmd in (init, grid, degrid, kclean, restore, model2comps, fluxtractor, hci, smoovie, sara):
    cli.add_command(cmd)

if __name__ == '__main__':
    cli()
